# Copyright (c) 2016 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import excutils

from cinder import context
from cinder import exception
from cinder.i18n import _
from cinder import utils
from cinder.volume.drivers.huawei import constants
from cinder.volume import qos_specs

LOG = logging.getLogger(__name__)


class SmartQos(object):
    def __init__(self, client):
        self.client = client

    @staticmethod
    def get_qos_by_volume_type(volume_type):
        # We prefer the qos_specs association
        # and override any existing extra-specs settings
        # if present.
        if not volume_type:
            return {}

        qos_specs_id = volume_type.get('qos_specs_id')
        if not qos_specs_id:
            return {}

        qos = {}
        io_type_flag = None
        ctxt = context.get_admin_context()
        consumer = qos_specs.get_qos_specs(ctxt, qos_specs_id)['consumer']
        if consumer == 'front-end':
            return {}

        kvs = qos_specs.get_qos_specs(ctxt, qos_specs_id)['specs']
        LOG.info('The QoS sepcs is: %s.', kvs)
        for k, v in kvs.items():
            if k not in constants.HUAWEI_VALID_KEYS:
                continue
            if k != 'IOType' and int(v) <= 0:
                msg = _('QoS config is wrong. %s must > 0.') % k
                LOG.error(msg)
                raise exception.InvalidInput(reason=msg)
            if k == 'IOType':
                if v not in ['0', '1', '2']:
                    msg = _('Illegal value specified for IOTYPE: 0, 1, or 2.')
                    LOG.error(msg)
                    raise exception.InvalidInput(reason=msg)
                io_type_flag = 1
                qos[k.upper()] = v
            else:
                qos[k.upper()] = v

        if not io_type_flag:
            msg = (_('QoS policy must specify for IOTYPE: 0, 1, or 2, '
                     'QoS policy: %(qos_policy)s ') % {'qos_policy': qos})
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

        # QoS policy must specify for IOTYPE and another qos_specs.
        if len(qos) < 2:
            msg = (_('QoS policy must specify for IOTYPE and another '
                     'qos_specs, QoS policy: %(qos_policy)s.')
                   % {'qos_policy': qos})
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

        for upper_limit in constants.UPPER_LIMIT_KEYS:
            for lower_limit in constants.LOWER_LIMIT_KEYS:
                if upper_limit in qos and lower_limit in qos:
                    msg = (_('QoS policy upper_limit and lower_limit '
                             'conflict, QoS policy: %(qos_policy)s.')
                           % {'qos_policy': qos})
                    LOG.error(msg)
                    raise exception.InvalidInput(reason=msg)

        return qos

    def _is_high_priority(self, qos):
        """Check QoS priority."""
        for key, value in qos.items():
            if (key.find('MIN') == 0) or (key.find('LATENCY') == 0):
                return True

        return False

    @utils.synchronized('huawei_qos', external=True)
    def add(self, qos, lun_id):
        policy_id = None
        try:
            # Check QoS priority.
            if self._is_high_priority(qos):
                self.client.change_lun_priority(lun_id)
            # Create QoS policy and activate it.
            version = self.client.find_array_version()
            if version >= constants.ARRAY_VERSION:
                (qos_id, lun_list) = self.client.find_available_qos(qos)
                if qos_id:
                    self.client.add_lun_to_qos(qos_id, lun_id, lun_list)
                else:
                    policy_id = self.client.create_qos_policy(qos, lun_id)
                    self.client.activate_deactivate_qos(policy_id, True)
            else:
                policy_id = self.client.create_qos_policy(qos, lun_id)
                self.client.activate_deactivate_qos(policy_id, True)
        except exception.VolumeBackendAPIException:
            with excutils.save_and_reraise_exception():
                if policy_id is not None:
                    self.client.delete_qos_policy(policy_id)

    @utils.synchronized('huawei_qos', external=True)
    def remove(self, qos_id, lun_id):
        qos_info = self.client.get_qos_info(qos_id)
        lun_list = self.client.get_lun_list_in_qos(qos_id, qos_info)
        if len(lun_list) <= 1:
            qos_status = qos_info['RUNNINGSTATUS']
            # 2: Active status.
            if qos_status != constants.STATUS_QOS_INACTIVE:
                self.client.activate_deactivate_qos(qos_id, False)
            self.client.delete_qos_policy(qos_id)
        else:
            self.client.remove_lun_from_qos(lun_id, lun_list, qos_id)


class SmartPartition(object):
    def __init__(self, client):
        self.client = client

    def add(self, opts, lun_id):
        if opts['smartpartition'] != 'true':
            return
        if not opts['partitionname']:
            raise exception.InvalidInput(
                reason=_('Partition name is None, please set '
                         'smartpartition:partitionname in key.'))

        partition_id = self.client.get_partition_id_by_name(
            opts['partitionname'])
        if not partition_id:
            raise exception.InvalidInput(
                reason=(_('Can not find partition id by name %(name)s.')
                        % {'name': opts['partitionname']}))

        self.client.add_lun_to_partition(lun_id, partition_id)


class SmartCache(object):
    def __init__(self, client):
        self.client = client

    def add(self, opts, lun_id):
        if opts['smartcache'] != 'true':
            return
        if not opts['cachename']:
            raise exception.InvalidInput(
                reason=_('Cache name is None, please set '
                         'smartcache:cachename in key.'))

        cache_id = self.client.get_cache_id_by_name(opts['cachename'])
        if not cache_id:
            raise exception.InvalidInput(
                reason=(_('Can not find cache id by cache name %(name)s.')
                        % {'name': opts['cachename']}))

        self.client.add_lun_to_cache(lun_id, cache_id)


class SmartX(object):
    def get_smartx_specs_opts(self, opts):
        # Check that smarttier is 0/1/2/3
        opts = self.get_smarttier_opts(opts)
        opts = self.get_smartthin_opts(opts)
        opts = self.get_smartcache_opts(opts)
        opts = self.get_smartpartition_opts(opts)
        return opts

    def get_smarttier_opts(self, opts):
        if opts['smarttier'] == 'true':
            if not opts['policy']:
                opts['policy'] = '1'
            elif opts['policy'] not in ['0', '1', '2', '3']:
                raise exception.InvalidInput(
                    reason=(_('Illegal value specified for smarttier: '
                              'set to either 0, 1, 2, or 3.')))
        else:
            opts['policy'] = '0'

        return opts

    def get_smartthin_opts(self, opts):
        if opts['thin_provisioning_support'] == 'true':
            if opts['thick_provisioning_support'] == 'true':
                raise exception.InvalidInput(
                    reason=(_('Illegal value specified for thin: '
                              'Can not set thin and thick at the same time.')))
            else:
                opts['LUNType'] = constants.THIN_LUNTYPE
        if opts['thick_provisioning_support'] == 'true':
            opts['LUNType'] = constants.THICK_LUNTYPE

        return opts

    def get_smartcache_opts(self, opts):
        if opts['smartcache'] == 'true':
            if not opts['cachename']:
                raise exception.InvalidInput(
                    reason=_('Cache name is None, please set '
                             'smartcache:cachename in key.'))
        else:
            opts['cachename'] = None

        return opts

    def get_smartpartition_opts(self, opts):
        if opts['smartpartition'] == 'true':
            if not opts['partitionname']:
                raise exception.InvalidInput(
                    reason=_('Partition name is None, please set '
                             'smartpartition:partitionname in key.'))
        else:
            opts['partitionname'] = None

        return opts



# Copyright (c) 2012 - 2016 EMC Corporation, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from cinder import context
from cinder.objects import fields
from cinder import test
from cinder.tests.unit import fake_constants as fake
from cinder.volume.drivers.coprhd import common as coprhd_common
from cinder.volume.drivers.coprhd import fc as coprhd_fc
from cinder.volume.drivers.coprhd import iscsi as coprhd_iscsi
from cinder.volume.drivers.coprhd import scaleio as coprhd_scaleio
from cinder.volume import volume_types

"""
Test Data required for mocking
"""
export_group_details_data = {
    "inactive": False,
    "initiators": [{"creation_time": 1392194176020,
                    "host": {"id": "urn:storageos:Host:3e21edff-8662-4e60-ab5",
                             "link": {"href": "/compute/hosts/urn:storageos:H",
                                      "rel": "self"}},
                    "hostname": "lglw7134",
                    "id": "urn:storageos:Initiator:13945431-06b7-44a0-838c-50",
                    "inactive": False,
                    "initiator_node": "20:00:00:90:FA:13:81:8D",
                    "initiator_port": "iqn.1993-08.org.deb:01:222",
                    "link": {"href": "/compute/initiators/urn:storageos:Initi",
                             "rel": "self"},
                    "protocol": "iSCSI",
                    "registration_status": "REGISTERED",
                    "tags": []}],
    "name": "ccgroup",
    "project": 'project',
    "tags": [],
    "tenant": 'tenant',
    "type": "Host",
    "varray": {"id": "urn:storageos:VirtualArray:5af376e9-ce2f-493d-9079-a872",
               "link": {"href": "/vdc/varrays/urn:storageos:VirtualArray:5af3",
                        "rel": "self"}
               },
    "volumes": [{"id": "urn:storageos:Volume:6dc64865-bb25-431c-b321-ac268f16"
                 "a7ae:vdc1",
                 "lun": 1
                 }]
}

varray_detail_data = {"name": "varray"}

export_group_list = ["urn:storageos:ExportGroup:2dbce233-7da0-47cb-8ff3-68f48"]

iscsi_itl_list = {"itl": [{"hlu": 3,
                           "initiator": {"id": "urn:storageos:Initiator:13945",
                                         "link": {"rel": "self",
                                                  "href": "/comput"},
                                         "port": "iqn.1993-08.org.deb:01:222"},
                           "export": {"id": "urn:storageos:ExportGroup:2dbce2",
                                      "name": "ccgroup",
                                      "link": {"rel": "self",
                                               "href": "/block/expo"}},
                           "device": {"id": "urn:storageos:Volume:aa1fc84a-af",
                                      "link": {"rel": "self",
                                               "href": "/block/volumes/urn:s"},
                                      "wwn": "600009700001957015735330303535"},
                           "target": {"id": "urn:storageos:StoragePort:d7e42",
                                      "link": {"rel": "self",
                                               "href": "/vdc/stor:"},
                                      "port": "50:00:09:73:00:18:95:19",
                                      'ip_address': "10.10.10.10",
                                      'tcp_port': '22'}},
                          {"hlu": 3,
                           "initiator": {"id": "urn:storageos:Initiator:13945",
                                         "link": {"rel": "self",
                                                  "href": "/comput"},
                                         "port": "iqn.1993-08.org.deb:01:222"},
                           "export": {"id": "urn:storageos:ExportGroup:2dbce2",
                                      "name": "ccgroup",
                                      "link": {"rel": "self",
                                               "href": "/block/expo"}},
                           "device": {"id": "urn:storageos:Volume:aa1fc84a-af",
                                      "link": {"rel": "self",
                                               "href": "/block/volumes/urn:s"},
                                      "wwn": "600009700001957015735330303535"},
                           "target": {"id": "urn:storageos:StoragePort:d7e42",
                                      "link": {"rel": "self",
                                               "href": "/vdc/stor:"},
                                      "port": "50:00:09:73:00:18:95:19",
                                      'ip_address': "10.10.10.10",
                                      'tcp_port': '22'}}]}

fcitl_itl_list = {"itl": [{"hlu": 3,
                           "initiator": {"id": "urn:storageos:Initiator:13945",
                                         "link": {"rel": "self",
                                                  "href": "/comput"},
                                         "port": "12:34:56:78:90:12:34:56"},
                           "export": {"id": "urn:storageos:ExportGroup:2dbce2",
                                      "name": "ccgroup",
                                      "link": {"rel": "self",
                                               "href": "/block/expo"}},
                           "device": {"id": "urn:storageos:Volume:aa1fc84a-af",
                                      "link": {"rel": "self",
                                               "href": "/block/volumes/urn:s"},
                                      "wwn": "600009700001957015735330303535"},
                           "target": {"id": "urn:storageos:StoragePort:d7e42",
                                      "link": {"rel": "self",
                                               "href": "/vdc/stor:"},
                                      "port": "12:34:56:78:90:12:34:56",
                                      'ip_address': "10.10.10.10",
                                      'tcp_port': '22'}},
                          {"hlu": 3,
                           "initiator": {"id": "urn:storageos:Initiator:13945",
                                         "link": {"rel": "self",
                                                  "href": "/comput"},
                                         "port": "12:34:56:78:90:12:34:56"},
                           "export": {"id": "urn:storageos:ExportGroup:2dbce2",
                                      "name": "ccgroup",
                                      "link": {"rel": "self",
                                               "href": "/block/expo"}},
                           "device": {"id": "urn:storageos:Volume:aa1fc84a-af",
                                      "link": {"rel": "self",
                                               "href": "/block/volumes/urn:s"},
                                      "wwn": "600009700001957015735330303535"},
                           "target": {"id": "urn:storageos:StoragePort:d7e42",
                                      "link": {"rel": "self",
                                               "href": "/vdc/stor:"},
                                      "port": "12:34:56:78:90:12:34:56",
                                      'ip_address': "10.10.10.10",
                                      'tcp_port': '22'}}]}

scaleio_itl_list = {"itl": [{"hlu": -1,
                             "initiator": {"id":
                                           "urn:storageos:Initiator:920aee",
                                           "link": {"rel": "self",
                                                    "href":
                                                    "/compute/initiators"},
                                           "port": "bfdf432500000004"},
                             "export": {"id":
                                        "urn:storageos:ExportGroup:5449235",
                                        "name": "10.108.225.109",
                                        "link": {"rel": "self",
                                                 "href":
                                                 "/block/exports/urn:stor"}},
                             "device": {"id":
                                        "urn:storageos:Volume:b3624a83-3eb",
                                        "link": {"rel": "self",
                                                 "href": "/block/volume"},
                                        "wwn":
                                        "4F48CC4C27A43248092128B400000004"},
                             "target": {}},
                            {"hlu": -1,
                             "initiator": {"id":
                                           "urn:storageos:Initiator:920aee",
                                           "link": {"rel": "self",
                                                    "href":
                                                    "/compute/initiators/"},
                                           "port": "bfdf432500000004"},
                             "export": {"id":
                                        "urn:storageos:ExportGroup:5449235",
                                        "name": "10.108.225.109",
                                        "link": {"rel": "self",
                                                 "href":
                                                 "/block/exports/urn:stor"}},
                             "device": {"id":
                                        "urn:storageos:Volume:c014e96a-557",
                                        "link": {"rel": "self",
                                                 "href":
                                                 "/block/volumes/urn:stor"},
                                        "wwn":
                                        "4F48CC4C27A43248092129320000000E"},
                             "target": {}}]}


class test_volume_data(object):
    name = 'test-vol1'
    size = 1
    volume_name = 'test-vol1'
    id = fake.VOLUME_ID
    group_id = None
    provider_auth = None
    project_id = fake.PROJECT_ID
    display_name = 'test-vol1'
    display_description = 'test volume',
    volume_type_id = None
    provider_id = fake.PROVIDER_ID

    def __init__(self, volume_type_id):
        self.volume_type_id = volume_type_id


class source_test_volume_data(object):
    name = 'source_test-vol1'
    size = 1
    volume_name = 'source_test-vol1'
    id = fake.VOLUME2_ID
    group_id = None
    provider_auth = None
    project_id = fake.PROJECT_ID
    display_name = 'source_test-vol1'
    display_description = 'test volume'
    volume_type_id = None

    def __init__(self, volume_type_id):
        self.volume_type_id = volume_type_id


class test_clone_volume_data(object):
    name = 'clone-test-vol1'
    size = 1
    volume_name = 'clone-test-vol1'
    id = fake.VOLUME3_ID
    provider_auth = None
    project_id = fake.PROJECT_ID
    display_name = 'clone-test-vol1'
    display_description = 'clone test volume'
    volume_type_id = None

    def __init__(self, volume_type_id):
        self.volume_type_id = volume_type_id


class test_snapshot_data(object):
    name = 'snapshot1'
    display_name = 'snapshot1'
    size = 1
    id = fake.SNAPSHOT_ID
    volume_name = 'test-vol1'
    volume_id = fake.VOLUME_ID
    volume = None
    volume_size = 1
    project_id = fake.PROJECT_ID
    status = fields.SnapshotStatus.AVAILABLE

    def __init__(self, src_volume):
        self.volume = src_volume


def get_connector_data():
    connector = {'ip': '10.0.0.2',
                 'initiator': 'iqn.1993-08.org.deb:01:222',
                 'wwpns': ["1234567890123456", "1234567890543211"],
                 'wwnns': ["223456789012345", "223456789054321"],
                 'host': 'fakehost'}
    return connector


class test_group_data(object):
    name = 'group_name'
    display_name = 'group_name'
    id = fake.GROUP_ID
    volume_type_ids = None
    volume_types = None
    group_type_id = None
    status = fields.GroupStatus.AVAILABLE

    def __init__(self, volume_types, group_type_id):
        self.group_type_id = group_type_id
        self.volume_types = volume_types


class test_group_type_data(object):
    name = 'group_name'
    display_name = 'group_name'
    groupsnapshot_id = None
    id = fake.GROUP_TYPE_ID
    description = 'group'


class test_group_snap_data(object):
    name = 'cg_snap_name'
    display_name = 'cg_snap_name'
    id = fake.GROUP_SNAPSHOT_ID
    group_id = fake.GROUP_ID
    status = fields.GroupStatus.AVAILABLE
    snapshots = []
    group = None
    group_type_id = None

    def __init__(self, volume_types, group_type_id):
        self.group_type_id = group_type_id
        self.group = test_group_data(volume_types, group_type_id)


class MockedEMCCoprHDDriverCommon(coprhd_common.EMCCoprHDDriverCommon):

    def __init__(self, protocol, default_backend_name,
                 configuration=None):

        super(MockedEMCCoprHDDriverCommon, self).__init__(
            protocol, default_backend_name, configuration)

    def authenticate_user(self):
        pass

    def get_exports_count_by_initiators(self, initiator_ports):
        return 0

    def _get_coprhd_volume_name(self, vol, verbose=False):
        if verbose is True:
            return {'volume_name': "coprhd_vol_name",
                    'volume_uri': "coprhd_vol_uri"}
        else:
            return "coprhd_vol_name"

    def _get_coprhd_snapshot_name(self, snapshot, resUri):
        return "coprhd_snapshot_name"

    def _get_coprhd_cgid(self, cgid):
        return "cg_uri"

    def init_volume_api(self):
        self.volume_api = mock.Mock()
        self.volume_api.get.return_value = {
            'name': 'source_test-vol1',
            'size': 1,
            'volume_name': 'source_test-vol1',
            'id': fake.VOLUME_ID,
            'group_id': fake.GROUP_ID,
            'provider_auth': None,
            'project_id': fake.PROJECT_ID,
            'display_name': 'source_test-vol1',
            'display_description': 'test volume',
            'volume_type_id': fake.VOLUME_TYPE_ID}

    def init_coprhd_api_components(self):
        self.volume_obj = mock.Mock()
        self.volume_obj.create.return_value = "volume_created"
        self.volume_obj.volume_query.return_value = "volume_uri"
        self.volume_obj.get_storageAttributes.return_value = (
            'block', 'volume_name')
        self.volume_obj.storage_resource_query.return_value = "volume_uri"
        self.volume_obj.is_volume_detachable.return_value = False
        self.volume_obj.volume_clone_detach.return_value = 'detached'
        self.volume_obj.getTags.return_value = (
            ["Openstack-vol", "Openstack-vol1"])
        self.volume_obj.tag.return_value = "tagged"
        self.volume_obj.clone.return_value = "volume-cloned"

        if(self.protocol == "iSCSI"):
            self.volume_obj.get_exports_by_uri.return_value = (
                iscsi_itl_list)
        elif(self.protocol == "FC"):
            self.volume_obj.get_exports_by_uri.return_value = (
                fcitl_itl_list)
        else:
            self.volume_obj.get_exports_by_uri.return_value = (
                scaleio_itl_list)

        self.volume_obj.list_volumes.return_value = []
        self.volume_obj.show.return_value = {"id": "vol_id"}
        self.volume_obj.expand.return_value = "expanded"

        self.tag_obj = mock.Mock()
        self.tag_obj.list_tags.return_value = [
            "Openstack-vol", "Openstack-vol1"]
        self.tag_obj.tag_resource.return_value = "Tagged"

        self.exportgroup_obj = mock.Mock()
        self.exportgroup_obj.exportgroup_list.return_value = (
            export_group_list)
        self.exportgroup_obj.exportgroup_show.return_value = (
            export_group_details_data)

        self.exportgroup_obj.exportgroup_add_volumes.return_value = (
            "volume-added")

        self.host_obj = mock.Mock()
        self.host_obj.list_by_tenant.return_value = []
        self.host_obj.list_all.return_value = [{'id': "host1_id",
                                                'name': "host1"}]
        self.host_obj.list_initiators.return_value = [
            {'name': "12:34:56:78:90:12:34:56"},
            {'name': "12:34:56:78:90:54:32:11"},
            {'name': "bfdf432500000004"}]

        self.hostinitiator_obj = mock.Mock()
        self.varray_obj = mock.Mock()
        self.varray_obj.varray_show.return_value = varray_detail_data

        self.snapshot_obj = mock.Mock()
        mocked_snap_obj = self.snapshot_obj.return_value
        mocked_snap_obj.storageResource_query.return_value = (
            "resourceUri")
        mocked_snap_obj.snapshot_create.return_value = (
            "snapshot_created")
        mocked_snap_obj.snapshot_query.return_value = "snapshot_uri"

        self.consistencygroup_obj = mock.Mock()
        mocked_group_object = self.consistencygroup_obj.return_value
        mocked_group_object.create.return_value = "CG-Created"
        mocked_group_object.consistencygroup_query.return_value = "CG-uri"


class EMCCoprHDISCSIDriverTest(test.TestCase):

    def setUp(self):
        super(EMCCoprHDISCSIDriverTest, self).setUp()
        self.create_coprhd_setup()

    def create_coprhd_setup(self):

        self.configuration = mock.Mock()
        self.configuration.coprhd_hostname = "10.10.10.10"
        self.configuration.coprhd_port = "4443"
        self.configuration.volume_backend_name = "EMCCoprHDISCSIDriver"
        self.configuration.coprhd_username = "user-name"
        self.configuration.coprhd_password = "password"
        self.configuration.coprhd_tenant = "tenant"
        self.configuration.coprhd_project = "project"
        self.configuration.coprhd_varray = "varray"
        self.configuration.coprhd_emulate_snapshot = False

        self.volume_type = self.create_coprhd_volume_type()
        self.volume_type_id = self.volume_type.id
        self.group_type = test_group_type_data()
        self.group_type_id = self.group_type.id

        self.mock_object(coprhd_iscsi.EMCCoprHDISCSIDriver,
                         '_get_common_driver',
                         self._get_mocked_common_driver)
        self.driver = coprhd_iscsi.EMCCoprHDISCSIDriver(
            configuration=self.configuration)

    def tearDown(self):
        self._cleanUp()
        super(EMCCoprHDISCSIDriverTest, self).tearDown()

    def _cleanUp(self):
        self.delete_vipr_volume_type()

    def create_coprhd_volume_type(self):
        ctx = context.get_admin_context()
        vipr_volume_type = volume_types.create(ctx,
                                               "coprhd-volume-type",
                                               {'CoprHD:VPOOL':
                                                'vpool_coprhd'})
        return vipr_volume_type

    def _get_mocked_common_driver(self):
        return MockedEMCCoprHDDriverCommon(
            protocol="iSCSI",
            default_backend_name="EMCViPRISCSIDriver",
            configuration=self.configuration)

    def delete_vipr_volume_type(self):
        ctx = context.get_admin_context()
        volume_types.destroy(ctx, self.volume_type_id)

    def test_create_destroy(self):
        volume = test_volume_data(self.volume_type_id)

        self.driver.create_volume(volume)
        self.driver.delete_volume(volume)

    def test_get_volume_stats(self):
        vol_stats = self.driver.get_volume_stats(True)
        self.assertEqual('unknown', vol_stats['free_capacity_gb'])

    def test_create_volume_clone(self):
        src_volume_data = test_volume_data(self.volume_type_id)
        clone_volume_data = test_clone_volume_data(self.volume_type_id)
        self.driver.create_volume(src_volume_data)
        self.driver.create_cloned_volume(clone_volume_data, src_volume_data)
        self.driver.delete_volume(src_volume_data)
        self.driver.delete_volume(clone_volume_data)

    def test_create_destroy_snapshot(self):
        volume_data = test_volume_data(self.volume_type_id)
        snapshot_data = test_snapshot_data(
            source_test_volume_data(self.volume_type_id))

        self.driver.create_volume(volume_data)
        self.driver.create_snapshot(snapshot_data)
        self.driver.delete_snapshot(snapshot_data)
        self.driver.delete_volume(volume_data)

    def test_create_volume_from_snapshot(self):

        src_vol_data = source_test_volume_data(self.volume_type_id)
        self.driver.create_volume(src_vol_data)

        volume_data = test_volume_data(self.volume_type_id)
        snapshot_data = test_snapshot_data(src_vol_data)

        self.driver.create_snapshot(snapshot_data)
        self.driver.create_volume_from_snapshot(volume_data, snapshot_data)

        self.driver.delete_snapshot(snapshot_data)
        self.driver.delete_volume(src_vol_data)
        self.driver.delete_volume(volume_data)

    def test_extend_volume(self):
        volume_data = test_volume_data(self.volume_type_id)
        self.driver.create_volume(volume_data)
        self.driver.extend_volume(volume_data, 2)
        self.driver.delete_volume(volume_data)

    def test_initialize_and_terminate_connection(self):
        connector_data = get_connector_data()
        volume_data = test_volume_data(self.volume_type_id)

        self.driver.create_volume(volume_data)
        res_initialize = self.driver.initialize_connection(
            volume_data, connector_data)
        expected_initialize = {'driver_volume_type': 'iscsi',
                               'data': {'target_lun': 3,
                                        'target_portal': '10.10.10.10:22',
                                        'target_iqn':
                                        '50:00:09:73:00:18:95:19',
                                        'target_discovered': False,
                                        'volume_id': fake.VOLUME_ID}}
        self.assertEqual(
            expected_initialize, res_initialize, 'Unexpected return data')

        self.driver.terminate_connection(volume_data, connector_data)
        self.driver.delete_volume(volume_data)

    @mock.patch('cinder.volume.utils.is_group_a_cg_snapshot_type')
    def test_create_delete_empty_group(self, cg_ss_enabled):
        cg_ss_enabled.side_effect = [True, True]
        group_data = test_group_data([self.volume_type],
                                     self.group_type_id)
        ctx = context.get_admin_context()
        self.driver.create_group(ctx, group_data)
        model_update, volumes_model_update = (
            self.driver.delete_group(ctx, group_data, []))
        self.assertEqual([], volumes_model_update, 'Unexpected return data')

    @mock.patch('cinder.volume.utils.is_group_a_cg_snapshot_type')
    def test_create_update_delete_group(self, cg_ss_enabled):
        cg_ss_enabled.side_effect = [True, True, True, True]
        group_data = test_group_data([self.volume_type],
                                     self.group_type_id)
        ctx = context.get_admin_context()
        self.driver.create_group(ctx, group_data)

        volume = test_volume_data(self.volume_type_id)
        self.driver.create_volume(volume)

        model_update, ret1, ret2 = (
            self.driver.update_group(ctx, group_data, [volume], []))

        self.assertEqual({'status': fields.GroupStatus.AVAILABLE},
                         model_update)

        model_update, volumes_model_update = (
            self.driver.delete_group(ctx, group_data, [volume]))
        self.assertEqual({'status': fields.GroupStatus.AVAILABLE},
                         model_update)
        self.assertEqual([{'status': 'deleted', 'id': fake.VOLUME_ID}],
                         volumes_model_update)

    @mock.patch('cinder.volume.utils.is_group_a_cg_snapshot_type')
    def test_create_delete_group_snap(self, cg_ss_enabled):
        cg_ss_enabled.side_effect = [True, True]
        group_snap_data = test_group_snap_data([self.volume_type],
                                               self.group_type_id)
        ctx = context.get_admin_context()

        model_update, snapshots_model_update = (
            self.driver.create_group_snapshot(ctx, group_snap_data, []))
        self.assertEqual({'status': fields.GroupStatus.AVAILABLE},
                         model_update)
        self.assertEqual([], snapshots_model_update, 'Unexpected return data')

        model_update, snapshots_model_update = (
            self.driver.delete_group_snapshot(ctx, group_snap_data, []))
        self.assertEqual({}, model_update, 'Unexpected return data')
        self.assertEqual([], snapshots_model_update, 'Unexpected return data')


class EMCCoprHDFCDriverTest(test.TestCase):

    def setUp(self):
        super(EMCCoprHDFCDriverTest, self).setUp()
        self.create_coprhd_setup()

    def create_coprhd_setup(self):

        self.configuration = mock.Mock()
        self.configuration.coprhd_hostname = "10.10.10.10"
        self.configuration.coprhd_port = "4443"
        self.configuration.volume_backend_name = "EMCCoprHDFCDriver"
        self.configuration.coprhd_username = "user-name"
        self.configuration.coprhd_password = "password"
        self.configuration.coprhd_tenant = "tenant"
        self.configuration.coprhd_project = "project"
        self.configuration.coprhd_varray = "varray"
        self.configuration.coprhd_emulate_snapshot = False

        self.volume_type = self.create_coprhd_volume_type()
        self.volume_type_id = self.volume_type.id
        self.group_type = test_group_type_data()
        self.group_type_id = self.group_type.id

        self.mock_object(coprhd_fc.EMCCoprHDFCDriver,
                         '_get_common_driver',
                         self._get_mocked_common_driver)
        self.driver = coprhd_fc.EMCCoprHDFCDriver(
            configuration=self.configuration)

    def tearDown(self):
        self._cleanUp()
        super(EMCCoprHDFCDriverTest, self).tearDown()

    def _cleanUp(self):
        self.delete_vipr_volume_type()

    def create_coprhd_volume_type(self):
        ctx = context.get_admin_context()
        vipr_volume_type = volume_types.create(ctx,
                                               "coprhd-volume-type",
                                               {'CoprHD:VPOOL': 'vpool_vipr'})
        return vipr_volume_type

    def _get_mocked_common_driver(self):
        return MockedEMCCoprHDDriverCommon(
            protocol="FC",
            default_backend_name="EMCViPRFCDriver",
            configuration=self.configuration)

    def delete_vipr_volume_type(self):
        ctx = context.get_admin_context()
        volume_types.destroy(ctx, self.volume_type_id)

    def test_create_destroy(self):
        volume = test_volume_data(self.volume_type_id)

        self.driver.create_volume(volume)
        self.driver.delete_volume(volume)

    def test_get_volume_stats(self):
        vol_stats = self.driver.get_volume_stats(True)
        self.assertEqual('unknown', vol_stats['free_capacity_gb'])

    def test_create_volume_clone(self):

        src_volume_data = test_volume_data(self.volume_type_id)
        clone_volume_data = test_clone_volume_data(self.volume_type_id)
        self.driver.create_volume(src_volume_data)
        self.driver.create_cloned_volume(clone_volume_data, src_volume_data)
        self.driver.delete_volume(src_volume_data)
        self.driver.delete_volume(clone_volume_data)

    def test_create_destroy_snapshot(self):

        volume_data = test_volume_data(self.volume_type_id)
        snapshot_data = test_snapshot_data(
            source_test_volume_data(self.volume_type_id))

        self.driver.create_volume(volume_data)
        self.driver.create_snapshot(snapshot_data)
        self.driver.delete_snapshot(snapshot_data)
        self.driver.delete_volume(volume_data)

    def test_create_volume_from_snapshot(self):
        src_vol_data = source_test_volume_data(self.volume_type_id)
        self.driver.create_volume(src_vol_data)

        volume_data = test_volume_data(self.volume_type_id)
        snapshot_data = test_snapshot_data(src_vol_data)

        self.driver.create_snapshot(snapshot_data)
        self.driver.create_volume_from_snapshot(volume_data, snapshot_data)

        self.driver.delete_snapshot(snapshot_data)
        self.driver.delete_volume(src_vol_data)
        self.driver.delete_volume(volume_data)

    def test_extend_volume(self):
        volume_data = test_volume_data(self.volume_type_id)
        self.driver.create_volume(volume_data)
        self.driver.extend_volume(volume_data, 2)
        self.driver.delete_volume(volume_data)

    def test_initialize_and_terminate_connection(self):

        connector_data = get_connector_data()
        volume_data = test_volume_data(self.volume_type_id)

        self.driver.create_volume(volume_data)
        res_initiatlize = self.driver.initialize_connection(
            volume_data, connector_data)
        expected_initialize = {'driver_volume_type': 'fibre_channel',
                               'data': {'target_lun': 3,
                                        'initiator_target_map':
                                        {'1234567890543211':
                                         ['1234567890123456',
                                          '1234567890123456'],
                                         '1234567890123456':
                                         ['1234567890123456',
                                          '1234567890123456']},
                                        'target_wwn': ['1234567890123456',
                                                       '1234567890123456'],
                                        'target_discovered': False,
                                        'volume_id': fake.VOLUME_ID}}
        self.assertEqual(
            expected_initialize, res_initiatlize, 'Unexpected return data')

        res_terminate = self.driver.terminate_connection(
            volume_data, connector_data)
        expected_terminate = {'driver_volume_type': 'fibre_channel',
                              'data': {'initiator_target_map':
                                       {'1234567890543211':
                                        ['1234567890123456',
                                         '1234567890123456'],
                                        '1234567890123456':
                                        ['1234567890123456',
                                         '1234567890123456']},
                                       'target_wwn': ['1234567890123456',
                                                      '1234567890123456']}}
        self.assertEqual(
            expected_terminate, res_terminate, 'Unexpected return data')

        self.driver.delete_volume(volume_data)

    @mock.patch('cinder.volume.utils.is_group_a_cg_snapshot_type')
    def test_create_delete_empty_group(self, cg_ss_enabled):
        cg_ss_enabled.side_effect = [True, True]
        group_data = test_group_data([self.volume_type],
                                     self.group_type_id)
        ctx = context.get_admin_context()
        self.driver.create_group(ctx, group_data)
        model_update, volumes_model_update = (
            self.driver.delete_group(ctx, group_data, []))
        self.assertEqual([], volumes_model_update, 'Unexpected return data')

    @mock.patch('cinder.volume.utils.is_group_a_cg_snapshot_type')
    def test_create_update_delete_group(self, cg_ss_enabled):
        cg_ss_enabled.side_effect = [True, True, True]
        group_data = test_group_data([self.volume_type],
                                     self.group_type_id)
        ctx = context.get_admin_context()
        self.driver.create_group(ctx, group_data)

        volume = test_volume_data(self.volume_type_id)
        self.driver.create_volume(volume)

        model_update, ret1, ret2 = (
            self.driver.update_group(ctx, group_data, [volume], []))

        self.assertEqual({'status': fields.GroupStatus.AVAILABLE},
                         model_update)

        model_update, volumes_model_update = (
            self.driver.delete_group(ctx, group_data, [volume]))
        self.assertEqual({'status': fields.GroupStatus.AVAILABLE},
                         model_update)
        self.assertEqual([{'status': 'deleted', 'id': fake.VOLUME_ID}],
                         volumes_model_update)

    @mock.patch('cinder.volume.utils.is_group_a_cg_snapshot_type')
    def test_create_delete_group_snap(self, cg_ss_enabled):
        cg_ss_enabled.side_effect = [True, True]
        group_snap_data = test_group_snap_data([self.volume_type],
                                               self.group_type_id)
        ctx = context.get_admin_context()

        model_update, snapshots_model_update = (
            self.driver.create_group_snapshot(ctx, group_snap_data, []))
        self.assertEqual({'status': fields.GroupStatus.AVAILABLE},
                         model_update)
        self.assertEqual([], snapshots_model_update, 'Unexpected return data')

        model_update, snapshots_model_update = (
            self.driver.delete_group_snapshot(ctx, group_snap_data, []))
        self.assertEqual({}, model_update, 'Unexpected return data')
        self.assertEqual([], snapshots_model_update, 'Unexpected return data')


class EMCCoprHDScaleIODriverTest(test.TestCase):

    def setUp(self):
        super(EMCCoprHDScaleIODriverTest, self).setUp()
        self.create_coprhd_setup()

    def create_coprhd_setup(self):

        self.configuration = mock.Mock()
        self.configuration.coprhd_hostname = "10.10.10.10"
        self.configuration.coprhd_port = "4443"
        self.configuration.volume_backend_name = "EMCCoprHDFCDriver"
        self.configuration.coprhd_username = "user-name"
        self.configuration.coprhd_password = "password"
        self.configuration.coprhd_tenant = "tenant"
        self.configuration.coprhd_project = "project"
        self.configuration.coprhd_varray = "varray"
        self.configuration.coprhd_scaleio_rest_gateway_host = "10.10.10.11"
        self.configuration.coprhd_scaleio_rest_gateway_port = 443
        self.configuration.coprhd_scaleio_rest_server_username = (
            "scaleio_username")
        self.configuration.coprhd_scaleio_rest_server_password = (
            "scaleio_password")
        self.configuration.scaleio_verify_server_certificate = False
        self.configuration.scaleio_server_certificate_path = (
            "/etc/scaleio/certs")

        self.volume_type = self.create_coprhd_volume_type()
        self.volume_type_id = self.volume_type.id
        self.group_type = test_group_type_data()
        self.group_type_id = self.group_type.id

        self.mock_object(coprhd_scaleio.EMCCoprHDScaleIODriver,
                         '_get_common_driver',
                         self._get_mocked_common_driver)
        self.mock_object(coprhd_scaleio.EMCCoprHDScaleIODriver,
                         '_get_client_id',
                         self._get_client_id)
        self.driver = coprhd_scaleio.EMCCoprHDScaleIODriver(
            configuration=self.configuration)

    def tearDown(self):
        self._cleanUp()
        super(EMCCoprHDScaleIODriverTest, self).tearDown()

    def _cleanUp(self):
        self.delete_vipr_volume_type()

    def create_coprhd_volume_type(self):
        ctx = context.get_admin_context()
        vipr_volume_type = volume_types.create(ctx,
                                               "coprhd-volume-type",
                                               {'CoprHD:VPOOL': 'vpool_vipr'})
        return vipr_volume_type

    def _get_mocked_common_driver(self):
        return MockedEMCCoprHDDriverCommon(
            protocol="scaleio",
            default_backend_name="EMCCoprHDScaleIODriver",
            configuration=self.configuration)

    def _get_client_id(self, server_ip, server_port, server_username,
                       server_password, sdc_ip):
        return "bfdf432500000004"

    def delete_vipr_volume_type(self):
        ctx = context.get_admin_context()
        volume_types.destroy(ctx, self.volume_type_id)

    def test_create_destroy(self):
        volume = test_volume_data(self.volume_type_id)

        self.driver.create_volume(volume)
        self.driver.delete_volume(volume)

    def test_get_volume_stats(self):
        vol_stats = self.driver.get_volume_stats(True)
        self.assertEqual('unknown', vol_stats['free_capacity_gb'])

    def test_create_volume_clone(self):

        src_volume_data = test_volume_data(self.volume_type_id)
        clone_volume_data = test_clone_volume_data(self.volume_type_id)
        self.driver.create_volume(src_volume_data)
        self.driver.create_cloned_volume(clone_volume_data, src_volume_data)
        self.driver.delete_volume(src_volume_data)
        self.driver.delete_volume(clone_volume_data)

    def test_create_destroy_snapshot(self):

        volume_data = test_volume_data(self.volume_type_id)
        snapshot_data = test_snapshot_data(
            source_test_volume_data(self.volume_type_id))

        self.driver.create_volume(volume_data)
        self.driver.create_snapshot(snapshot_data)
        self.driver.delete_snapshot(snapshot_data)
        self.driver.delete_volume(volume_data)

    def test_create_volume_from_snapshot(self):
        src_vol_data = source_test_volume_data(self.volume_type_id)
        self.driver.create_volume(src_vol_data)

        volume_data = test_volume_data(self.volume_type_id)
        snapshot_data = test_snapshot_data(src_vol_data)

        self.driver.create_snapshot(snapshot_data)
        self.driver.create_volume_from_snapshot(volume_data, snapshot_data)

        self.driver.delete_snapshot(snapshot_data)
        self.driver.delete_volume(src_vol_data)
        self.driver.delete_volume(volume_data)

    def test_extend_volume(self):
        volume_data = test_volume_data(self.volume_type_id)
        self.driver.create_volume(volume_data)
        self.driver.extend_volume(volume_data, 2)
        self.driver.delete_volume(volume_data)

    def test_initialize_and_terminate_connection(self):

        connector_data = get_connector_data()
        volume_data = test_volume_data(self.volume_type_id)

        self.driver.create_volume(volume_data)
        res_initiatlize = self.driver.initialize_connection(
            volume_data, connector_data)
        exp_name = res_initiatlize['data']['scaleIO_volname']
        expected_initialize = {'data': {'bandwidthLimit': None,
                                        'hostIP': '10.0.0.2',
                                        'iopsLimit': None,
                                        'scaleIO_volname': exp_name,
                                        'scaleIO_volume_id': fake.PROVIDER_ID,
                                        'serverIP': '10.10.10.11',
                                        'serverPassword': 'scaleio_password',
                                        'serverPort': 443,
                                        'serverToken': None,
                                        'serverUsername': 'scaleio_username'},
                               'driver_volume_type': 'scaleio'}
        self.assertEqual(
            expected_initialize, res_initiatlize, 'Unexpected return data')

        self.driver.terminate_connection(
            volume_data, connector_data)
        self.driver.delete_volume(volume_data)

    @mock.patch('cinder.volume.utils.is_group_a_cg_snapshot_type')
    def test_create_delete_empty_group(self, cg_ss_enabled):
        cg_ss_enabled.side_effect = [True, True]
        group_data = test_group_data([self.volume_type],
                                     self.group_type_id)
        ctx = context.get_admin_context()
        self.driver.create_group(ctx, group_data)
        model_update, volumes_model_update = (
            self.driver.delete_group(ctx, group_data, []))
        self.assertEqual([], volumes_model_update, 'Unexpected return data')

    @mock.patch('cinder.volume.utils.is_group_a_cg_snapshot_type')
    def test_create_update_delete_group(self, cg_ss_enabled):
        cg_ss_enabled.side_effect = [True, True, True, True]
        group_data = test_group_data([self.volume_type],
                                     self.group_type_id)
        ctx = context.get_admin_context()
        self.driver.create_group(ctx, group_data)

        volume = test_volume_data(self.volume_type_id)
        self.driver.create_volume(volume)

        model_update, ret1, ret2 = (
            self.driver.update_group(ctx, group_data, [volume], []))

        self.assertEqual({'status': fields.GroupStatus.AVAILABLE},
                         model_update)

        model_update, volumes_model_update = (
            self.driver.delete_group(ctx, group_data, [volume]))
        self.assertEqual({'status': fields.GroupStatus.AVAILABLE},
                         model_update)
        self.assertEqual([{'status': 'deleted', 'id': fake.VOLUME_ID}],
                         volumes_model_update)

    @mock.patch('cinder.volume.utils.is_group_a_cg_snapshot_type')
    def test_create_delete_group_snap(self, cg_ss_enabled):
        cg_ss_enabled.side_effect = [True, True]
        group_snap_data = test_group_snap_data([self.volume_type],
                                               self.group_type_id)
        ctx = context.get_admin_context()

        model_update, snapshots_model_update = (
            self.driver.create_group_snapshot(ctx, group_snap_data, []))
        self.assertEqual({'status': fields.GroupStatus.AVAILABLE},
                         model_update)
        self.assertEqual([], snapshots_model_update, 'Unexpected return data')

        model_update, snapshots_model_update = (
            self.driver.delete_group_snapshot(ctx, group_snap_data, []))
        self.assertEqual({}, model_update, 'Unexpected return data')
        self.assertEqual([], snapshots_model_update, 'Unexpected return data')



#    copyright (c) 2016 Industrial Technology Research Institute.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Class for DISCO to attach and detach volume."""

from os_brick import initiator
from os_brick.initiator import connector
from oslo_log import log as logging

from cinder import utils


LOG = logging.getLogger(__name__)


class AttachDetachDiscoVolume(object):
    """Class for attach and detach a DISCO volume."""

    def __init__(self, configuration):
        """Init volume attachment class."""
        self.configuration = configuration
        self.connector = connector.InitiatorConnector.factory(
            self._get_connector_identifier(), utils.get_root_helper(),
            device_scan_attempts=(
                self.configuration.num_volume_device_scan_tries)
        )
        self.connection_conf = {}
        self.connection_conf['server_ip'] = self.configuration.disco_client
        self.connection_conf['server_port'] = (
            self.configuration.disco_client_port)

        self.connection_properties = {}
        self.connection_properties['name'] = None
        self.connection_properties['disco_id'] = None
        self.connection_properties['conf'] = self.connection_conf

    def _get_connection_properties(self, volume):
        """Return a dictionnary with the connection properties."""
        connection_properties = dict(self.connection_properties)
        connection_properties['name'] = volume['name']
        connection_properties['disco_id'] = volume['provider_location']
        return connection_properties

    def _get_connector_identifier(self):
        """Return connector identifier, put here to mock it in unit tests."""
        return initiator.DISCO

    def _attach_volume(self, volume):
        """Call the connector.connect_volume()."""
        connection_properties = self._get_connection_properties(volume)
        device_info = self.connector.connect_volume(connection_properties)
        return device_info

    def _detach_volume(self, volume):
        """Call the connector.disconnect_volume()."""
        connection_properties = self._get_connection_properties(volume)
        self.connector.disconnect_volume(connection_properties, volume)



# Copyright (c) 2017 Veritas Technologies LLC.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Error Codes
"""

EX_BAD_PARAM = 10
EX_BAD_MESSAGE = 106
MSG_SUCCESS = 0
MSG_ERROR = 1

"""
Constants
"""
HS_VHOST = "/"
ACK_YES = 1
ACK_NO = 0
BLK_YES = 1
BLK_NO = 0
EXCH_DIRECT = "direct"
EXCH_FANOUT = "fanout"
EXCH_TOPIC = "topic"

MSG_REQUEST = 1
MSG_RESPONSE = 2
MSG_TOKEN = "token"
MSG_OWNER = "owner"
MSG_TYPE = "type"
MSG_ERROR = "err_code"
MSG_ACK = "ack"
MSG_BLK = "blocking"
MSG_BLK_INFO = "blocking_info"
MSG_BLK_NAME = "name"
MSG_BLK_BINDKEY = "bindkey"
MSG_BLK_TYPE = "type"
MSG_PAYLOAD = "payload"

# HyperScale Controller Exchange
HS_CONTROLLER_EXCH = 'hyperscale-controller'
HS_RPC_EXCH = 'hyperscale-recv'
HS_DATANODE_EXCH = 'hyperscale-datanode'
HS_COMPUTE_EXCH = 'hyperscale-storage'

SNAP_RESTORE_RF = 3



# Copyright (c) 2015 Alex Meade.  All rights reserved.
# Copyright (c) 2015 Michael Price.  All rights reserved.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import copy
import ddt
import mock
import socket

from cinder import exception
from cinder.volume import configuration as conf

from cinder.tests.unit.volume.drivers.netapp.eseries import fakes as \
    fakes
from cinder.volume.drivers.netapp import common
from cinder.volume.drivers.netapp.eseries import client
from cinder.volume.drivers.netapp.eseries import library
from cinder.volume.drivers.netapp.eseries import utils
from cinder.volume.drivers.netapp import options
import cinder.volume.drivers.netapp.utils as na_utils


@ddt.ddt
class NetAppESeriesDriverTestCase(object):
    """Test case for NetApp e-series iscsi driver."""

    volume = {'id': '114774fb-e15a-4fae-8ee2-c9723e3645ef', 'size': 1,
              'volume_name': 'lun1', 'host': 'hostname@backend#DDP',
              'os_type': 'linux', 'provider_location': 'lun1',
              'name_id': '114774fb-e15a-4fae-8ee2-c9723e3645ef',
              'provider_auth': 'provider a b', 'project_id': 'project',
              'display_name': None, 'display_description': 'lun1',
              'volume_type_id': None}
    snapshot = {'id': '17928122-553b-4da9-9737-e5c3dcd97f75',
                'volume_id': '114774fb-e15a-4fae-8ee2-c9723e3645ef',
                'size': 2, 'volume_name': 'lun1',
                'volume_size': 2, 'project_id': 'project',
                'display_name': None, 'display_description': 'lun1',
                'volume_type_id': None}
    volume_sec = {'id': 'b6c01641-8955-4917-a5e3-077147478575',
                  'size': 2, 'volume_name': 'lun1',
                  'os_type': 'linux', 'provider_location': 'lun1',
                  'name_id': 'b6c01641-8955-4917-a5e3-077147478575',
                  'provider_auth': None, 'project_id': 'project',
                  'display_name': None, 'display_description': 'lun1',
                  'volume_type_id': None}
    volume_clone = {'id': 'b4b24b27-c716-4647-b66d-8b93ead770a5', 'size': 3,
                    'volume_name': 'lun1',
                    'os_type': 'linux', 'provider_location': 'cl_sm',
                    'name_id': 'b4b24b27-c716-4647-b66d-8b93ead770a5',
                    'provider_auth': None,
                    'project_id': 'project', 'display_name': None,
                    'display_description': 'lun1',
                    'volume_type_id': None}
    volume_clone_large = {'id': 'f6ef5bf5-e24f-4cbb-b4c4-11d631d6e553',
                          'size': 6, 'volume_name': 'lun1',
                          'os_type': 'linux', 'provider_location': 'cl_lg',
                          'name_id': 'f6ef5bf5-e24f-4cbb-b4c4-11d631d6e553',
                          'provider_auth': None,
                          'project_id': 'project', 'display_name': None,
                          'display_description': 'lun1',
                          'volume_type_id': None}
    fake_eseries_volume_label = utils.convert_uuid_to_es_fmt(volume['id'])
    fake_size_gb = volume['size']
    fake_eseries_pool_label = 'DDP'
    fake_ref = {'source-name': 'CFDGJSLS'}
    fake_ret_vol = {'id': 'vol_id', 'label': 'label',
                    'worldWideName': 'wwn', 'capacity': '2147583648'}
    PROTOCOL = 'iscsi'

    def setUp(self):
        super(NetAppESeriesDriverTestCase, self).setUp()
        self._custom_setup()

    def _custom_setup(self):
        self.mock_object(na_utils, 'OpenStackInfo')

        configuration = self._set_config(self.create_configuration())
        self.driver = common.NetAppDriver(configuration=configuration)
        self.library = self.driver.library
        self.mock_object(self.library,
                         '_check_mode_get_or_register_storage_system')
        self.mock_object(self.library, '_version_check')
        self.mock_object(self.driver.library, '_check_storage_system')
        self.driver.do_setup(context='context')
        self.driver.library._client._endpoint = fakes.FAKE_ENDPOINT_HTTP
        self.driver.library._client.features = mock.Mock()
        self.driver.library._client.features.REST_1_4_RELEASE = True

    def _set_config(self, configuration):
        configuration.netapp_storage_family = 'eseries'
        configuration.netapp_storage_protocol = self.PROTOCOL
        configuration.netapp_transport_type = 'http'
        configuration.netapp_server_hostname = '127.0.0.1'
        configuration.netapp_server_port = None
        configuration.netapp_webservice_path = '/devmgr/vn'
        configuration.netapp_controller_ips = '127.0.0.2,127.0.0.3'
        configuration.netapp_sa_password = 'pass1234'
        configuration.netapp_login = 'rw'
        configuration.netapp_password = 'rw'
        configuration.netapp_storage_pools = 'DDP'
        configuration.netapp_enable_multiattach = False
        return configuration

    @staticmethod
    def create_configuration():
        configuration = conf.Configuration(None)
        configuration.append_config_values(options.netapp_basicauth_opts)
        configuration.append_config_values(options.netapp_eseries_opts)
        configuration.append_config_values(options.netapp_san_opts)
        return configuration

    @abc.abstractmethod
    @mock.patch.object(na_utils, 'validate_instantiation')
    def test_instantiation(self, mock_validate_instantiation):
        pass

    def test_embedded_mode(self):
        self.mock_object(client.RestClient, '_init_features')
        configuration = self._set_config(self.create_configuration())
        configuration.netapp_controller_ips = '127.0.0.1,127.0.0.3'
        driver = common.NetAppDriver(configuration=configuration)
        self.mock_object(driver.library, '_version_check')
        self.mock_object(client.RestClient, 'list_storage_systems',
                         return_value=[fakes.STORAGE_SYSTEM])
        driver.do_setup(context='context')

        self.assertEqual('1fa6efb5-f07b-4de4-9f0e-52e5f7ff5d1b',
                         driver.library._client.get_system_id())

    def test_check_system_pwd_not_sync(self):
        def list_system():
            if getattr(self, 'test_count', None):
                self.test_count = 1
                return {'status': 'passwordoutofsync'}
            return {'status': 'needsAttention'}

        self.library._client.list_storage_system = mock.Mock(wraps=list_system)
        result = self.library._check_storage_system()
        self.assertTrue(bool(result))

    def test_create_destroy(self):
        self.mock_object(client.RestClient, 'delete_volume',
                         return_value='None')
        self.mock_object(self.driver.library, 'create_volume',
                         return_value=self.volume)
        self.mock_object(self.library._client, 'list_volume',
                         return_value=fakes.VOLUME)

        self.driver.create_volume(self.volume)
        self.driver.delete_volume(self.volume)

    def test_vol_stats(self):
        self.driver.get_volume_stats(refresh=False)

    def test_get_pool(self):
        self.mock_object(self.library, '_get_volume',
                         return_value={'volumeGroupRef': 'fake_ref'})
        self.mock_object(self.library._client, "get_storage_pool",
                         return_value={'volumeGroupRef': 'fake_ref',
                                       'label': 'ddp1'})

        pool = self.driver.get_pool({'name_id': 'fake-uuid'})

        self.assertEqual('ddp1', pool)

    def test_get_pool_no_pools(self):
        self.mock_object(self.library, '_get_volume',
                         return_value={'volumeGroupRef': 'fake_ref'})
        self.mock_object(self.library._client, "get_storage_pool",
                         return_value=None)

        pool = self.driver.get_pool({'name_id': 'fake-uuid'})

        self.assertIsNone(pool)

    @mock.patch.object(library.NetAppESeriesLibrary, '_create_volume',
                       mock.Mock())
    def test_create_volume(self):

        self.driver.create_volume(self.volume)

        self.library._create_volume.assert_called_with(
            'DDP', self.fake_eseries_volume_label, self.volume['size'], {})

    def test_create_volume_no_pool_provided_by_scheduler(self):
        volume = copy.deepcopy(self.volume)
        volume['host'] = "host@backend"  # missing pool
        self.assertRaises(exception.InvalidHost, self.driver.create_volume,
                          volume)

    @mock.patch.object(client.RestClient, 'list_storage_pools')
    def test_helper_create_volume_fail(self, fake_list_pools):
        fake_pool = {}
        fake_pool['label'] = self.fake_eseries_pool_label
        fake_pool['volumeGroupRef'] = 'foo'
        fake_pool['raidLevel'] = 'raidDiskPool'
        fake_pools = [fake_pool]
        fake_list_pools.return_value = fake_pools
        wrong_eseries_pool_label = 'hostname@backend'
        self.assertRaises(exception.NetAppDriverException,
                          self.library._create_volume,
                          wrong_eseries_pool_label,
                          self.fake_eseries_volume_label,
                          self.fake_size_gb)

    @mock.patch.object(library.LOG, 'info')
    @mock.patch.object(client.RestClient, 'list_storage_pools')
    @mock.patch.object(client.RestClient, 'create_volume',
                       mock.MagicMock(return_value='CorrectVolume'))
    def test_helper_create_volume(self, storage_pools, log_info):
        fake_pool = {}
        fake_pool['label'] = self.fake_eseries_pool_label
        fake_pool['volumeGroupRef'] = 'foo'
        fake_pool['raidLevel'] = 'raidDiskPool'
        fake_pools = [fake_pool]
        storage_pools.return_value = fake_pools
        storage_vol = self.library._create_volume(
            self.fake_eseries_pool_label,
            self.fake_eseries_volume_label,
            self.fake_size_gb)
        log_info.assert_called_once_with("Created volume with label %s.",
                                         self.fake_eseries_volume_label)
        self.assertEqual('CorrectVolume', storage_vol)

    @mock.patch.object(client.RestClient, 'list_storage_pools')
    @mock.patch.object(client.RestClient, 'create_volume',
                       mock.MagicMock(
                           side_effect=exception.NetAppDriverException))
    @mock.patch.object(library.LOG, 'info', mock.Mock())
    def test_create_volume_check_exception(self, fake_list_pools):
        fake_pool = {}
        fake_pool['label'] = self.fake_eseries_pool_label
        fake_pool['volumeGroupRef'] = 'foo'
        fake_pool['raidLevel'] = 'raidDiskPool'
        fake_pools = [fake_pool]
        fake_list_pools.return_value = fake_pools
        self.assertRaises(exception.NetAppDriverException,
                          self.library._create_volume,
                          self.fake_eseries_pool_label,
                          self.fake_eseries_volume_label, self.fake_size_gb)

    def test_portal_for_vol_controller(self):
        volume = {'id': 'vol_id', 'currentManager': 'ctrl1'}
        vol_nomatch = {'id': 'vol_id', 'currentManager': 'ctrl3'}
        portals = [{'controller': 'ctrl2', 'iqn': 'iqn2'},
                   {'controller': 'ctrl1', 'iqn': 'iqn1'}]
        portal = self.library._get_iscsi_portals_for_vol(volume, portals)
        self.assertEqual({'controller': 'ctrl1', 'iqn': 'iqn1'}, portal[0])
        portal = self.library._get_iscsi_portals_for_vol(vol_nomatch, portals)
        self.assertEqual({'controller': 'ctrl2', 'iqn': 'iqn2'}, portal[0])

    def test_portal_for_vol_any_false(self):
        vol_nomatch = {'id': 'vol_id', 'currentManager': 'ctrl3'}
        portals = [{'controller': 'ctrl2', 'iqn': 'iqn2'},
                   {'controller': 'ctrl1', 'iqn': 'iqn1'}]
        self.assertRaises(exception.NetAppDriverException,
                          self.library._get_iscsi_portals_for_vol,
                          vol_nomatch, portals, False)

    def test_do_setup_all_default(self):
        configuration = self._set_config(self.create_configuration())
        driver = common.NetAppDriver(configuration=configuration)
        driver.library._check_mode_get_or_register_storage_system = mock.Mock()
        mock_invoke = self.mock_object(client, 'RestClient')
        driver.do_setup(context='context')
        mock_invoke.assert_called_with(**fakes.FAKE_CLIENT_PARAMS)

    def test_do_setup_http_default_port(self):
        configuration = self._set_config(self.create_configuration())
        configuration.netapp_transport_type = 'http'
        driver = common.NetAppDriver(configuration=configuration)
        driver.library._check_mode_get_or_register_storage_system = mock.Mock()
        mock_invoke = self.mock_object(client, 'RestClient')
        driver.do_setup(context='context')
        mock_invoke.assert_called_with(**fakes.FAKE_CLIENT_PARAMS)

    def test_do_setup_https_default_port(self):
        configuration = self._set_config(self.create_configuration())
        configuration.netapp_transport_type = 'https'
        driver = common.NetAppDriver(configuration=configuration)
        driver.library._check_mode_get_or_register_storage_system = mock.Mock()
        mock_invoke = self.mock_object(client, 'RestClient')
        driver.do_setup(context='context')
        FAKE_EXPECTED_PARAMS = dict(fakes.FAKE_CLIENT_PARAMS, port=8443,
                                    scheme='https')
        mock_invoke.assert_called_with(**FAKE_EXPECTED_PARAMS)

    def test_do_setup_http_non_default_port(self):
        configuration = self._set_config(self.create_configuration())
        configuration.netapp_server_port = 81
        driver = common.NetAppDriver(configuration=configuration)
        driver.library._check_mode_get_or_register_storage_system = mock.Mock()
        mock_invoke = self.mock_object(client, 'RestClient')
        driver.do_setup(context='context')
        FAKE_EXPECTED_PARAMS = dict(fakes.FAKE_CLIENT_PARAMS, port=81)
        mock_invoke.assert_called_with(**FAKE_EXPECTED_PARAMS)

    def test_do_setup_https_non_default_port(self):
        configuration = self._set_config(self.create_configuration())
        configuration.netapp_transport_type = 'https'
        configuration.netapp_server_port = 446
        driver = common.NetAppDriver(configuration=configuration)
        driver.library._check_mode_get_or_register_storage_system = mock.Mock()
        mock_invoke = self.mock_object(client, 'RestClient')
        driver.do_setup(context='context')
        FAKE_EXPECTED_PARAMS = dict(fakes.FAKE_CLIENT_PARAMS, port=446,
                                    scheme='https')
        mock_invoke.assert_called_with(**FAKE_EXPECTED_PARAMS)

    def test_setup_good_controller_ip(self):
        configuration = self._set_config(self.create_configuration())
        configuration.netapp_controller_ips = '127.0.0.1'
        driver = common.NetAppDriver(configuration=configuration)
        driver.library._check_mode_get_or_register_storage_system

    def test_setup_good_controller_ips(self):
        configuration = self._set_config(self.create_configuration())
        configuration.netapp_controller_ips = '127.0.0.2,127.0.0.1'
        driver = common.NetAppDriver(configuration=configuration)
        driver.library._check_mode_get_or_register_storage_system

    def test_setup_missing_controller_ip(self):
        configuration = self._set_config(self.create_configuration())
        configuration.netapp_controller_ips = None
        driver = common.NetAppDriver(configuration=configuration)
        self.assertRaises(exception.InvalidInput,
                          driver.do_setup, context='context')

    def test_setup_error_invalid_controller_ip(self):
        configuration = self._set_config(self.create_configuration())
        configuration.netapp_controller_ips = '987.65.43.21'
        driver = common.NetAppDriver(configuration=configuration)
        self.mock_object(na_utils, 'resolve_hostname',
                         side_effect=socket.gaierror)

        self.assertRaises(
            exception.NoValidBackend,
            driver.library._check_mode_get_or_register_storage_system)

    def test_setup_error_invalid_first_controller_ip(self):
        configuration = self._set_config(self.create_configuration())
        configuration.netapp_controller_ips = '987.65.43.21,127.0.0.1'
        driver = common.NetAppDriver(configuration=configuration)
        self.mock_object(na_utils, 'resolve_hostname',
                         side_effect=socket.gaierror)

        self.assertRaises(
            exception.NoValidBackend,
            driver.library._check_mode_get_or_register_storage_system)

    def test_setup_error_invalid_second_controller_ip(self):
        configuration = self._set_config(self.create_configuration())
        configuration.netapp_controller_ips = '127.0.0.1,987.65.43.21'
        driver = common.NetAppDriver(configuration=configuration)
        self.mock_object(na_utils, 'resolve_hostname',
                         side_effect=socket.gaierror)

        self.assertRaises(
            exception.NoValidBackend,
            driver.library._check_mode_get_or_register_storage_system)

    def test_setup_error_invalid_both_controller_ips(self):
        configuration = self._set_config(self.create_configuration())
        configuration.netapp_controller_ips = '564.124.1231.1,987.65.43.21'
        driver = common.NetAppDriver(configuration=configuration)
        self.mock_object(na_utils, 'resolve_hostname',
                         side_effect=socket.gaierror)

        self.assertRaises(
            exception.NoValidBackend,
            driver.library._check_mode_get_or_register_storage_system)

    def test_manage_existing_get_size(self):
        self.library._get_existing_vol_with_manage_ref = mock.Mock(
            return_value=self.fake_ret_vol)
        size = self.driver.manage_existing_get_size(self.volume, self.fake_ref)
        self.assertEqual(3, size)
        self.library._get_existing_vol_with_manage_ref.assert_called_once_with(
            self.fake_ref)

    def test_get_exist_vol_source_name_missing(self):
        self.library._client.list_volume = mock.Mock(
            side_effect=exception.InvalidInput)
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.library._get_existing_vol_with_manage_ref,
                          {'id': '1234'})

    @ddt.data('source-id', 'source-name')
    def test_get_exist_vol_source_not_found(self, attr_name):
        def _get_volume(v_id):
            d = {'id': '1', 'name': 'volume1', 'worldWideName': '0'}
            if v_id in d:
                return d[v_id]
            else:
                raise exception.VolumeNotFound(message=v_id)

        self.library._client.list_volume = mock.Mock(wraps=_get_volume)
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.library._get_existing_vol_with_manage_ref,
                          {attr_name: 'name2'})

        self.library._client.list_volume.assert_called_once_with(
            'name2')

    def test_get_exist_vol_with_manage_ref(self):
        fake_ret_vol = {'id': 'right'}
        self.library._client.list_volume = mock.Mock(return_value=fake_ret_vol)

        actual_vol = self.library._get_existing_vol_with_manage_ref(
            {'source-name': 'name2'})

        self.library._client.list_volume.assert_called_once_with('name2')
        self.assertEqual(fake_ret_vol, actual_vol)

    @mock.patch.object(utils, 'convert_uuid_to_es_fmt')
    def test_manage_existing_same_label(self, mock_convert_es_fmt):
        self.library._get_existing_vol_with_manage_ref = mock.Mock(
            return_value=self.fake_ret_vol)
        mock_convert_es_fmt.return_value = 'label'
        self.driver.manage_existing(self.volume, self.fake_ref)
        self.library._get_existing_vol_with_manage_ref.assert_called_once_with(
            self.fake_ref)
        mock_convert_es_fmt.assert_called_once_with(
            '114774fb-e15a-4fae-8ee2-c9723e3645ef')

    @mock.patch.object(utils, 'convert_uuid_to_es_fmt')
    def test_manage_existing_new(self, mock_convert_es_fmt):
        self.library._get_existing_vol_with_manage_ref = mock.Mock(
            return_value=self.fake_ret_vol)
        mock_convert_es_fmt.return_value = 'vol_label'
        self.library._client.update_volume = mock.Mock(
            return_value={'id': 'update', 'worldWideName': 'wwn'})
        self.driver.manage_existing(self.volume, self.fake_ref)
        self.library._get_existing_vol_with_manage_ref.assert_called_once_with(
            self.fake_ref)
        mock_convert_es_fmt.assert_called_once_with(
            '114774fb-e15a-4fae-8ee2-c9723e3645ef')
        self.library._client.update_volume.assert_called_once_with(
            'vol_id', 'vol_label')

    @mock.patch.object(library.LOG, 'info')
    def test_unmanage(self, log_info):
        self.library._get_volume = mock.Mock(return_value=self.fake_ret_vol)
        self.driver.unmanage(self.volume)
        self.library._get_volume.assert_called_once_with(
            '114774fb-e15a-4fae-8ee2-c9723e3645ef')
        self.assertEqual(1, log_info.call_count)

    @mock.patch.object(library.NetAppESeriesLibrary, 'ensure_export',
                       mock.Mock())
    def test_ensure_export(self):
        self.driver.ensure_export('context', self.fake_ret_vol)
        self.assertTrue(self.library.ensure_export.called)

    @mock.patch.object(library.NetAppESeriesLibrary, 'extend_volume',
                       mock.Mock())
    def test_extend_volume(self):
        capacity = 10
        self.driver.extend_volume(self.fake_ret_vol, capacity)
        self.library.extend_volume.assert_called_with(self.fake_ret_vol,
                                                      capacity)

    @mock.patch.object(library.NetAppESeriesLibrary,
                       'create_cgsnapshot', mock.Mock())
    def test_create_cgsnapshot(self):
        cgsnapshot = copy.deepcopy(fakes.FAKE_CINDER_CG_SNAPSHOT)
        snapshots = copy.deepcopy([fakes.SNAPSHOT_IMAGE])

        self.driver.create_cgsnapshot('ctx', cgsnapshot, snapshots)

        self.library.create_cgsnapshot.assert_called_with(cgsnapshot,
                                                          snapshots)

    @mock.patch.object(library.NetAppESeriesLibrary,
                       'delete_cgsnapshot', mock.Mock())
    def test_delete_cgsnapshot(self):
        cgsnapshot = copy.deepcopy(fakes.FAKE_CINDER_CG_SNAPSHOT)
        snapshots = copy.deepcopy([fakes.SNAPSHOT_IMAGE])

        self.driver.delete_cgsnapshot('ctx', cgsnapshot, snapshots)

        self.library.delete_cgsnapshot.assert_called_with(cgsnapshot,
                                                          snapshots)

    @mock.patch.object(library.NetAppESeriesLibrary,
                       'create_consistencygroup', mock.Mock())
    def test_create_consistencygroup(self):
        cg = copy.deepcopy(fakes.FAKE_CINDER_CG)

        self.driver.create_consistencygroup('ctx', cg)

        self.library.create_consistencygroup.assert_called_with(cg)

    @mock.patch.object(library.NetAppESeriesLibrary,
                       'delete_consistencygroup', mock.Mock())
    def test_delete_consistencygroup(self):
        cg = copy.deepcopy(fakes.FAKE_CINDER_CG)
        volumes = copy.deepcopy([fakes.VOLUME])

        self.driver.delete_consistencygroup('ctx', cg, volumes)

        self.library.delete_consistencygroup.assert_called_with(cg, volumes)

    @mock.patch.object(library.NetAppESeriesLibrary,
                       'update_consistencygroup', mock.Mock())
    def test_update_consistencygroup(self):
        group = copy.deepcopy(fakes.FAKE_CINDER_CG)

        self.driver.update_consistencygroup('ctx', group, {}, {})

        self.library.update_consistencygroup.assert_called_with(group, {}, {})

    @mock.patch.object(library.NetAppESeriesLibrary,
                       'create_consistencygroup_from_src', mock.Mock())
    def test_create_consistencygroup_from_src(self):
        cg = copy.deepcopy(fakes.FAKE_CINDER_CG)
        volumes = copy.deepcopy([fakes.VOLUME])
        source_vols = copy.deepcopy([fakes.VOLUME])
        cgsnapshot = copy.deepcopy(fakes.FAKE_CINDER_CG_SNAPSHOT)
        source_cg = copy.deepcopy(fakes.FAKE_CINDER_CG_SNAPSHOT)
        snapshots = copy.deepcopy([fakes.SNAPSHOT_IMAGE])

        self.driver.create_consistencygroup_from_src(
            'ctx', cg, volumes, cgsnapshot, snapshots, source_cg,
            source_vols)

        self.library.create_consistencygroup_from_src.assert_called_with(
            cg, volumes, cgsnapshot, snapshots, source_cg, source_vols)



# Copyright (c) 2017 DataCore Software Corp. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Classes to invoke DataCore SANsymphony API."""

import copy
import sys
import uuid

from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import importutils
import retrying
import six
import socket
import suds
from suds import client as suds_client
from suds import plugin
from suds.sax import attribute
from suds.sax import element
from suds import wsdl
from suds import wsse
from suds import xsd

from cinder.i18n import _
from cinder import utils as cinder_utils
from cinder.volume.drivers.datacore import exception as datacore_exceptions
from cinder.volume.drivers.datacore import utils as datacore_utils

websocket = importutils.try_import('websocket')


LOG = logging.getLogger(__name__)


class FaultDefinitionsFilter(plugin.DocumentPlugin):
    """Plugin to process the DataCore API WSDL document.

    The document plugin removes fault definitions for callback operations
    from the DataCore API WSDL.
    """

    def parsed(self, context):
        document = context.document
        tns = self._get_tns(document)

        message_qrefs = set()
        for message in self._get_wsdl_messages(document):
            message_qrefs.add((message.get('name'), tns[1]))

        bindings = self._get_wsdl_operation_bindings(document)

        for port_type in self._get_wsdl_port_types(document):
            for operation in self._get_wsdl_operations(port_type):
                self._filter_faults(
                    document, operation, bindings, message_qrefs, tns)

    @staticmethod
    def _get_tns(document):
        target_namespace = document.get('targetNamespace')
        prefix = document.findPrefix(target_namespace) or 'tns'
        return prefix, target_namespace

    @staticmethod
    def _get_wsdl_port_types(document):
        return document.getChildren('portType', wsdl.wsdlns)

    @staticmethod
    def _get_wsdl_operations(port_type):
        return port_type.getChildren('operation', wsdl.wsdlns)

    @staticmethod
    def _get_wsdl_messages(document):
        return document.getChildren('message', wsdl.wsdlns)

    @staticmethod
    def _get_wsdl_operation_bindings(document):
        bindings = []
        for binding in document.getChildren('binding', wsdl.wsdlns):
            operations = {}
            for operation in binding.getChildren('operation', wsdl.wsdlns):
                operations[operation.get('name')] = operation
            bindings.append(operations)
        return bindings

    @staticmethod
    def _filter_faults(document, operation, operation_bindings,
                       message_qrefs, tns):
        filtered_faults = {}
        for fault in operation.getChildren('fault', wsdl.wsdlns):
            fault_message = fault.get('message')
            qref = xsd.qualify(fault_message, document, tns)
            if qref not in message_qrefs:
                filtered_faults[fault.get('name')] = fault
        for fault in filtered_faults.values():
            operation.remove(fault)
        if filtered_faults:
            for binding in operation_bindings:
                filtered_binding_faults = []
                faults = binding[operation.get('name')].getChildren(
                    'fault', wsdl.wsdlns)
                for binding_fault in faults:
                    if binding_fault.get('name') in filtered_faults:
                        filtered_binding_faults.append(binding_fault)
                for binding_fault in filtered_binding_faults:
                    binding[operation.get('name')].remove(binding_fault)


class DataCoreClient(object):
    """DataCore SANsymphony client."""

    API_RETRY_INTERVAL = 10

    DATACORE_EXECUTIVE_PORT = '3794'

    STORAGE_SERVICES = 'IStorageServices'
    STORAGE_SERVICES_BINDING = 'CustomBinding_IStorageServices'

    EXECUTIVE_SERVICE = 'IExecutiveServiceEx'
    EXECUTIVE_SERVICE_BINDING = 'CustomBinding_IExecutiveServiceEx'

    NS_WSA = ('wsa', 'http://www.w3.org/2005/08/addressing')
    WSA_ANONYMOUS = 'http://www.w3.org/2005/08/addressing/anonymous'
    MUST_UNDERSTAND = attribute.Attribute('SOAP-ENV:mustUnderstand', '1')

    # Namespaces that are defined within DataCore API WSDL
    NS_DATACORE_EXECUTIVE = ('http://schemas.datacontract.org/2004/07/'
                             'DataCore.Executive')
    NS_DATACORE_EXECUTIVE_SCSI = ('http://schemas.datacontract.org/2004/07/'
                                  'DataCore.Executive.Scsi')
    NS_DATACORE_EXECUTIVE_ISCSI = ('http://schemas.datacontract.org/2004/07/'
                                   'DataCore.Executive.iSCSI')
    NS_SERIALIZATION_ARRAYS = ('http://schemas.microsoft.com/2003/10/'
                               'Serialization/Arrays')

    # Fully qualified names of objects that are defined within
    # DataCore API WSDL
    O_ACCESS_TOKEN = '{%s}AccessToken' % NS_DATACORE_EXECUTIVE_ISCSI
    O_ARRAY_OF_PERFORMANCE_TYPE = ('{%s}ArrayOfPerformanceType'
                                   % NS_DATACORE_EXECUTIVE)
    O_ARRAY_OF_STRING = '{%s}ArrayOfstring' % NS_SERIALIZATION_ARRAYS
    O_CLIENT_MACHINE_TYPE = '{%s}ClientMachineType' % NS_DATACORE_EXECUTIVE
    O_DATA_SIZE = '{%s}DataSize' % NS_DATACORE_EXECUTIVE
    O_LOGICAL_DISK_ROLE = '{%s}LogicalDiskRole' % NS_DATACORE_EXECUTIVE
    O_LOGICAL_UNIT_TYPE = '{%s}LogicalUnitType' % NS_DATACORE_EXECUTIVE
    O_MIRROR_RECOVERY_PRIORITY = ('{%s}MirrorRecoveryPriority'
                                  % NS_DATACORE_EXECUTIVE)
    O_PATH_POLICY = '{%s}PathPolicy' % NS_DATACORE_EXECUTIVE
    O_PERFORMANCE_TYPE = '{%s}PerformanceType' % NS_DATACORE_EXECUTIVE
    O_POOL_VOLUME_TYPE = '{%s}PoolVolumeType' % NS_DATACORE_EXECUTIVE
    O_SNAPSHOT_TYPE = '{%s}SnapshotType' % NS_DATACORE_EXECUTIVE
    O_SCSI_MODE = '{%s}ScsiMode' % NS_DATACORE_EXECUTIVE_SCSI
    O_SCSI_PORT_DATA = '{%s}ScsiPortData' % NS_DATACORE_EXECUTIVE
    O_SCSI_PORT_NEXUS_DATA = '{%s}ScsiPortNexusData' % NS_DATACORE_EXECUTIVE
    O_SCSI_PORT_TYPE = '{%s}ScsiPortType' % NS_DATACORE_EXECUTIVE_SCSI
    O_VIRTUAL_DISK_DATA = '{%s}VirtualDiskData' % NS_DATACORE_EXECUTIVE
    O_VIRTUAL_DISK_STATUS = '{%s}VirtualDiskStatus' % NS_DATACORE_EXECUTIVE
    O_VIRTUAL_DISK_SUB_TYPE = '{%s}VirtualDiskSubType' % NS_DATACORE_EXECUTIVE
    O_VIRTUAL_DISK_TYPE = '{%s}VirtualDiskType' % NS_DATACORE_EXECUTIVE

    def __init__(self, host, username, password, timeout):
        if websocket is None:
            msg = _("Failed to import websocket-client python module."
                    " Please, ensure the module is installed.")
            raise datacore_exceptions.DataCoreException(msg)

        self.timeout = timeout

        executive_service_net_addr = datacore_utils.build_network_address(
            host, self.DATACORE_EXECUTIVE_PORT)
        executive_service_endpoint = self._build_service_endpoint(
            executive_service_net_addr, self.EXECUTIVE_SERVICE)

        security_options = wsse.Security()
        username_token = wsse.UsernameToken(username, password)
        security_options.tokens.append(username_token)

        self._executive_service_client = suds_client.Client(
            executive_service_endpoint['http_endpoint'] + '?singlewsdl',
            nosend=True,
            timeout=self.timeout,
            wsse=security_options,
            plugins=[FaultDefinitionsFilter()])

        self._update_storage_services_endpoint(executive_service_endpoint)

        storage_services_endpoint = self._get_storage_services_endpoint()

        self._storage_services_client = suds_client.Client(
            storage_services_endpoint['http_endpoint'] + '?singlewsdl',
            nosend=True,
            timeout=self.timeout,
            wsse=security_options,
            plugins=[FaultDefinitionsFilter()])

        self._update_executive_service_endpoints(storage_services_endpoint)

    @staticmethod
    def _get_list_data(obj, attribute_name):
        return getattr(obj, attribute_name, [])

    @staticmethod
    def _build_service_endpoint(network_address, path):
        return {
            'network_address': network_address,
            'http_endpoint': '%s://%s/%s' % ('http', network_address, path),
            'ws_endpoint': '%s://%s/%s' % ('ws', network_address, path),
        }

    @cinder_utils.synchronized('datacore-api-request_context')
    def _get_soap_context(self, service_client, service_binding, method,
                          message_id, *args, **kwargs):
        soap_action = (service_client.wsdl.services[0].port(service_binding)
                       .methods[method].soap.action)

        soap_headers = self._get_soap_headers(soap_action, message_id)

        service_client.set_options(soapheaders=soap_headers)
        context = service_client.service[service_binding][method](
            *args, **kwargs)

        return context

    def _get_soap_headers(self, soap_action, message_id):
        headers = [
            element.Element('Action', ns=self.NS_WSA)
            .setText(soap_action.replace('"', ''))
            .append(self.MUST_UNDERSTAND),

            element.Element('To', ns=self.NS_WSA)
            .setText(self.WSA_ANONYMOUS)
            .append(self.MUST_UNDERSTAND),

            element.Element('MessageID', ns=self.NS_WSA)
            .setText(message_id),

            element.Element('ReplyTo', ns=self.NS_WSA)
            .insert(element.Element('Address', ns=self.NS_WSA)
                    .setText(self.WSA_ANONYMOUS)),
        ]
        return headers

    def _process_request(self, service_client, service_binding,
                         service_endpoint, method, *args, **kwargs):
        message_id = uuid.uuid4().urn

        context = self._get_soap_context(
            service_client, service_binding,
            method, message_id, *args, **kwargs)

        channel = None
        try:
            channel = websocket.create_connection(
                service_endpoint,
                timeout=self.timeout,
                subprotocols=['soap'],
                header=['soap-content-type: text/xml'])
            channel.send(context.envelope)
            response = channel.recv()
            if isinstance(response, six.text_type):
                response = response.encode('utf-8')
            return context.process_reply(response)
        except (socket.error, websocket.WebSocketException) as e:
            traceback = sys.exc_info()[2]
            error = datacore_exceptions.DataCoreConnectionException(reason=e)
            six.reraise(datacore_exceptions.DataCoreConnectionException,
                        error,
                        traceback)
        except suds.WebFault as e:
            traceback = sys.exc_info()[2]
            fault = datacore_exceptions.DataCoreFaultException(reason=e)
            six.reraise(datacore_exceptions.DataCoreFaultException,
                        fault,
                        traceback)
        finally:
            if channel and channel.connected:
                try:
                    channel.close()
                except (socket.error, websocket.WebSocketException) as e:
                    LOG.debug("Closing a connection to "
                              "DataCore server failed. %s", e)

    def _invoke_storage_services(self, method, *args, **kwargs):

        @retrying.retry(
            retry_on_exception=lambda e:
                isinstance(e, datacore_exceptions.DataCoreConnectionException),
            wait_fixed=self.API_RETRY_INTERVAL * 1000,
            stop_max_delay=self.timeout * 1000)
        def retry_call():
            storage_services_endpoint = self._get_storage_services_endpoint()
            try:
                result = self._process_request(
                    self._storage_services_client,
                    self.STORAGE_SERVICES_BINDING,
                    storage_services_endpoint['ws_endpoint'],
                    method, *args, **kwargs)
                return result
            except datacore_exceptions.DataCoreConnectionException:
                with excutils.save_and_reraise_exception():
                    self._update_api_endpoints()

        return retry_call()

    def _update_api_endpoints(self):
        executive_service_endpoints = self._get_executive_service_endpoints()
        for endpoint in executive_service_endpoints:
            try:
                self._update_storage_services_endpoint(endpoint)
                break
            except datacore_exceptions.DataCoreConnectionException as e:
                LOG.warning("Failed to update DataCore Server Group "
                            "endpoints. %s.", e)

        storage_services_endpoint = self._get_storage_services_endpoint()
        try:
            self._update_executive_service_endpoints(
                storage_services_endpoint)
        except datacore_exceptions.DataCoreConnectionException as e:
            LOG.warning("Failed to update DataCore Server Group "
                        "endpoints. %s.", e)

    @cinder_utils.synchronized('datacore-api-storage_services_endpoint')
    def _get_storage_services_endpoint(self):
        if self._storage_services_endpoint:
            return copy.copy(self._storage_services_endpoint)
        return None

    @cinder_utils.synchronized('datacore-api-storage_services_endpoint')
    def _update_storage_services_endpoint(self, executive_service_endpoint):
        controller_address = self._process_request(
            self._executive_service_client,
            self.EXECUTIVE_SERVICE_BINDING,
            executive_service_endpoint['ws_endpoint'],
            'GetControllerAddress')

        if not controller_address:
            msg = _("Could not determine controller node.")
            raise datacore_exceptions.DataCoreConnectionException(reason=msg)

        controller_host = controller_address.rsplit(':', 1)[0].strip('[]')
        controller_net_addr = datacore_utils.build_network_address(
            controller_host,
            self.DATACORE_EXECUTIVE_PORT)

        self._storage_services_endpoint = self._build_service_endpoint(
            controller_net_addr,
            self.STORAGE_SERVICES)

    @cinder_utils.synchronized('datacore-api-executive_service_endpoints')
    def _get_executive_service_endpoints(self):
        if self._executive_service_endpoints:
            return self._executive_service_endpoints[:]
        return []

    @cinder_utils.synchronized('datacore-api-executive_service_endpoints')
    def _update_executive_service_endpoints(self, storage_services_endpoint):
        endpoints = []
        nodes = self._get_list_data(
            self._process_request(self._storage_services_client,
                                  self.STORAGE_SERVICES_BINDING,
                                  storage_services_endpoint['ws_endpoint'],
                                  'GetNodes'),
            'RegionNodeData')

        if not nodes:
            msg = _("Could not determine executive nodes.")
            raise datacore_exceptions.DataCoreConnectionException(reason=msg)

        for node in nodes:
            host = node.HostAddress.rsplit(':', 1)[0].strip('[]')
            endpoint = self._build_service_endpoint(
                datacore_utils.build_network_address(
                    host, self.DATACORE_EXECUTIVE_PORT),
                self.EXECUTIVE_SERVICE)
            endpoints.append(endpoint)

        self._executive_service_endpoints = endpoints

    def get_server_groups(self):
        """Get all the server groups in the configuration.

        :return: A list of server group data.
        """

        return self._get_list_data(
            self._invoke_storage_services('GetServerGroups'),
            'ServerHostGroupData')

    def get_servers(self):
        """Get all the server hosts in the configuration.

        :return: A list of server host data
        """

        return self._get_list_data(
            self._invoke_storage_services('GetServers'),
            'ServerHostData')

    def get_disk_pools(self):
        """Get all the pools in the server group.

        :return: A list of disk pool data
        """

        return self._get_list_data(
            self._invoke_storage_services('GetDiskPools'),
            'DiskPoolData')

    def get_logical_disks(self):
        """Get all the logical disks defined in the system.

        :return: A list of logical disks
        """

        return self._get_list_data(
            self._invoke_storage_services('GetLogicalDisks'),
            'LogicalDiskData')

    def create_pool_logical_disk(self, pool_id, pool_volume_type, size,
                                 min_quota=None, max_quota=None):
        """Create the pool logical disk.

        :param pool_id: Pool id
        :param pool_volume_type: Type, either striped or spanned
        :param size: Size
        :param min_quota: Min quota
        :param max_quota: Max quota
        :return: New logical disk data
        """

        volume_type = getattr(self._storage_services_client.factory
                              .create(self.O_POOL_VOLUME_TYPE),
                              pool_volume_type)

        data_size = (self._storage_services_client.factory
                     .create(self.O_DATA_SIZE))
        data_size.Value = size

        data_size_min_quota = None
        if min_quota:
            data_size_min_quota = (self._storage_services_client.factory
                                   .create(self.O_DATA_SIZE))
            data_size_min_quota.Value = min_quota

        data_size_max_quota = None
        if max_quota:
            data_size_max_quota = (self._storage_services_client.factory
                                   .create(self.O_DATA_SIZE))
            data_size_max_quota.Value = max_quota

        return self._invoke_storage_services('CreatePoolLogicalDisk',
                                             poolId=pool_id,
                                             type=volume_type,
                                             size=data_size,
                                             minQuota=data_size_min_quota,
                                             maxQuota=data_size_max_quota)

    def delete_logical_disk(self, logical_disk_id):
        """Delete the logical disk.

        :param logical_disk_id: Logical disk id
        """

        self._invoke_storage_services('DeleteLogicalDisk',
                                      logicalDiskId=logical_disk_id)

    def get_logical_disk_chunk_allocation_map(self, logical_disk_id):
        """Get the logical disk chunk allocation map.

        The logical disk allocation map details all the physical disk chunks
        that are currently allocated to this logical disk.

        :param logical_disk_id: Logical disk id
        :return: A list of member allocation maps, restricted to chunks
                 allocated on to this logical disk
        """

        return self._get_list_data(
            self._invoke_storage_services('GetLogicalDiskChunkAllocationMap',
                                          logicalDiskId=logical_disk_id),
            'MemberAllocationInfoData')

    def get_next_virtual_disk_alias(self, base_alias):
        """Get the next available (unused) virtual disk alias.

        :param base_alias: Base string of the new alias
        :return: New alias
        """

        return self._invoke_storage_services('GetNextVirtualDiskAlias',
                                             baseAlias=base_alias)

    def get_virtual_disks(self):
        """Get all the virtual disks in the configuration.

        :return: A list of virtual disk's data
        """

        return self._get_list_data(
            self._invoke_storage_services('GetVirtualDisks'),
            'VirtualDiskData')

    def build_virtual_disk_data(self, virtual_disk_alias, virtual_disk_type,
                                size, description, storage_profile_id):
        """Create VirtualDiskData object.

        :param virtual_disk_alias: User-visible alias of the virtual disk,
                                   which must be unique
        :param virtual_disk_type: Virtual disk type
        :param size: Virtual disk size
        :param description: A user-readable description of the virtual disk
        :param storage_profile_id: Virtual disk storage profile
        :return: VirtualDiskData object
        """

        vd_data = (self._storage_services_client.factory
                   .create(self.O_VIRTUAL_DISK_DATA))
        vd_data.Size = (self._storage_services_client.factory
                        .create(self.O_DATA_SIZE))
        vd_data.Size.Value = size
        vd_data.Alias = virtual_disk_alias
        vd_data.Description = description
        vd_data.Type = getattr(self._storage_services_client.factory
                               .create(self.O_VIRTUAL_DISK_TYPE),
                               virtual_disk_type)
        vd_data.SubType = getattr(self._storage_services_client.factory
                                  .create(self.O_VIRTUAL_DISK_SUB_TYPE),
                                  'Standard')
        vd_data.DiskStatus = getattr(self._storage_services_client.factory
                                     .create(self.O_VIRTUAL_DISK_STATUS),
                                     'Online')
        vd_data.RecoveryPriority = getattr(
            self._storage_services_client.factory
            .create(self.O_MIRROR_RECOVERY_PRIORITY),
            'Unset')
        vd_data.StorageProfileId = storage_profile_id

        return vd_data

    def create_virtual_disk_ex2(self, virtual_disk_data, first_logical_disk_id,
                                second_logical_disk_id, add_redundancy):
        """Create a virtual disk specifying the both logical disks.

        :param virtual_disk_data: Virtual disk's properties
        :param first_logical_disk_id: Id of the logical disk to use
        :param second_logical_disk_id: Id of the second logical disk to use
        :param add_redundancy: If True, the mirror has redundant mirror paths
        :return: New virtual disk's data
        """

        return self._invoke_storage_services(
            'CreateVirtualDiskEx2',
            virtualDisk=virtual_disk_data,
            firstLogicalDiskId=first_logical_disk_id,
            secondLogicalDiskId=second_logical_disk_id,
            addRedundancy=add_redundancy)

    def set_virtual_disk_size(self, virtual_disk_id, size):
        """Change the size of a virtual disk.

        :param virtual_disk_id: Id of the virtual disk
        :param size: New size
        :return: Virtual disk's data
        """

        data_size = (self._storage_services_client.factory
                     .create(self.O_DATA_SIZE))
        data_size.Value = size

        return self._invoke_storage_services('SetVirtualDiskSize',
                                             virtualDiskId=virtual_disk_id,
                                             size=data_size)

    def delete_virtual_disk(self, virtual_disk_id, delete_logical_disks):
        """Delete a virtual disk.

        :param virtual_disk_id: Id of the virtual disk
        :param delete_logical_disks: If True, delete the associated
                                     logical disks
        """

        self._invoke_storage_services('DeleteVirtualDisk',
                                      virtualDiskId=virtual_disk_id,
                                      deleteLogicalDisks=delete_logical_disks)

    def serve_virtual_disks_to_host(self, host_id, virtual_disks):
        """Serve multiple virtual disks to a specified host.

        :param host_id: Id of the host machine
        :param virtual_disks: A list of virtual disks to serve
        :return: A list of the virtual disks actually served to the host
        """

        virtual_disk_array = (self._storage_services_client.factory
                              .create(self.O_ARRAY_OF_STRING))
        virtual_disk_array.string = virtual_disks

        return self._get_list_data(
            self._invoke_storage_services('ServeVirtualDisksToHost',
                                          hostId=host_id,
                                          virtualDisks=virtual_disk_array),
            'VirtualLogicalUnitData')

    def unserve_virtual_disks_from_host(self, host_id, virtual_disks):
        """Unserve multiple virtual disks from a specified host.

        :param host_id: Id of the host machine
        :param virtual_disks: A list of virtual disks to unserve
        """

        virtual_disk_array = (self._storage_services_client.factory
                              .create(self.O_ARRAY_OF_STRING))
        virtual_disk_array.string = virtual_disks

        self._invoke_storage_services('UnserveVirtualDisksFromHost',
                                      hostId=host_id,
                                      virtualDisks=virtual_disk_array)

    def unserve_virtual_disks_from_port(self, port_id, virtual_disks):
        """Unserve multiple virtual disks from a specified initiator port.

        :param port_id: Id of the initiator port
        :param virtual_disks: A list of virtual disks to unserve
        """

        virtual_disk_array = (self._storage_services_client.factory
                              .create(self.O_ARRAY_OF_STRING))
        virtual_disk_array.string = virtual_disks

        self._invoke_storage_services('UnserveVirtualDisksFromPort',
                                      portId=port_id,
                                      virtualDisks=virtual_disk_array)

    def bind_logical_disk(self, virtual_disk_id, logical_disk_id, role,
                          create_mirror_mappings, create_client_mappings,
                          add_redundancy):
        """Bind (add) a logical disk to a virtual disk.

        :param virtual_disk_id: Id of the virtual disk to bind to
        :param logical_disk_id: Id of the logical disk being bound
        :param role: logical disk's role
        :param create_mirror_mappings: If True, automatically create the
                                       mirror mappings to this disk, assuming
                                       there is already another logical disk
                                       bound
        :param create_client_mappings: If True, automatically create mappings
                                       from mapped hosts to the new disk
        :param add_redundancy: If True, the mirror has redundant mirror paths
        :return: Updated virtual disk data
        """

        logical_disk_role = getattr(self._storage_services_client.factory
                                    .create(self.O_LOGICAL_DISK_ROLE),
                                    role)

        return self._invoke_storage_services(
            'BindLogicalDisk',
            virtualDiskId=virtual_disk_id,
            logicalDiskId=logical_disk_id,
            role=logical_disk_role,
            createMirrorMappings=create_mirror_mappings,
            createClientMappings=create_client_mappings,
            addRedundancy=add_redundancy)

    def get_snapshots(self):
        """Get all the snapshots on all the servers in the region.

        :return: A list of snapshot data.
        """

        return self._get_list_data(
            self._invoke_storage_services('GetSnapshots'),
            'SnapshotData')

    def create_snapshot(self, virtual_disk_id, name, description,
                        destination_pool_id, snapshot_type,
                        duplicate_disk_id, storage_profile_id):
        """Create a snapshot relationship.

        :param virtual_disk_id: Virtual disk id
        :param name: Name of snapshot
        :param description: Description
        :param destination_pool_id: Destination pool id
        :param snapshot_type: Type of snapshot
        :param duplicate_disk_id: If set to True then the destination virtual
                                  disk's SCSI id will be a duplicate of the
                                  source's
        :param storage_profile_id: Specifies the destination virtual disk's
                                   storage profile
        :return: New snapshot data
        """

        st_type = getattr(self._storage_services_client.factory
                          .create(self.O_SNAPSHOT_TYPE),
                          snapshot_type)

        return self._invoke_storage_services(
            'CreateSnapshot',
            virtualDiskId=virtual_disk_id,
            name=name,
            description=description,
            destinationPoolId=destination_pool_id,
            type=st_type,
            duplicateDiskId=duplicate_disk_id,
            storageProfileId=storage_profile_id)

    def delete_snapshot(self, snapshot_id):
        """Delete the snapshot.

        :param snapshot_id: Snapshot id
        """

        self._invoke_storage_services('DeleteSnapshot', snapshotId=snapshot_id)

    def get_storage_profiles(self):
        """Get all the all the defined storage profiles.

        :return: A list of storage profiles
        """

        return self._get_list_data(
            self._invoke_storage_services('GetStorageProfiles'),
            'StorageProfileData')

    def designate_map_store(self, pool_id):
        """Designate which pool the snapshot mapstore will be allocated from.

        :param pool_id: Pool id
        :return: Updated server host data, which includes the mapstore pool id
        """

        return self._invoke_storage_services('DesignateMapStore',
                                             poolId=pool_id)

    def get_performance_by_type(self, performance_types):
        """Get performance data for specific types of performance counters.

        :param performance_types: A list of performance counter types
        :return: A list of performance data points
        """

        prfm_type_array = (self._storage_services_client.factory
                           .create(self.O_ARRAY_OF_PERFORMANCE_TYPE))
        prfm_type_array.PerformanceType = list(
            getattr(self._storage_services_client.factory
                    .create(self.O_PERFORMANCE_TYPE),
                    performance_type)
            for performance_type in performance_types)

        return self._get_list_data(
            self._invoke_storage_services('GetPerformanceByType',
                                          types=prfm_type_array),
            'CollectionPointData')

    def get_ports(self):
        """Get all ports in the configuration.

        :return: A list of SCSI ports
        """

        return self._get_list_data(
            self._invoke_storage_services('GetPorts'),
            'ScsiPortData')

    def build_scsi_port_data(self, host_id, port_name, port_mode, port_type):
        """Create ScsiPortData object that represents SCSI port, of any type.

        :param host_id: Id of the port's host computer
        :param port_name: Unique name of the port.
        :param port_mode: Mode of port: initiator or target
        :param port_type: Type of port, Fc, iSCSI or loopback
        :return: ScsiPortData object
        """

        scsi_port_data = (self._storage_services_client.factory
                          .create(self.O_SCSI_PORT_DATA))
        scsi_port_data.HostId = host_id
        scsi_port_data.PortName = port_name
        scsi_port_data.PortMode = getattr(self._storage_services_client.factory
                                          .create(self.O_SCSI_MODE),
                                          port_mode)
        scsi_port_data.PortType = getattr(self._storage_services_client.factory
                                          .create(self.O_SCSI_PORT_TYPE),
                                          port_type)

        return scsi_port_data

    def register_port(self, scsi_port_data):
        """Register a port in the configuration.

        :param scsi_port_data: Port data
        :return: Updated port data
        """

        return self._invoke_storage_services('RegisterPort',
                                             port=scsi_port_data)

    def assign_port(self, client_id, port_id):
        """Assign a port to a client.

        :param client_id: Client id
        :param port_id: Port id
        :return: Updated port data,
                 which will now have its host id set to the client id
        """

        return self._invoke_storage_services('AssignPort',
                                             clientId=client_id,
                                             portId=port_id)

    def set_server_port_properties(self, port_id, properties):
        """Set a server port's properties.

        :param port_id: Port id
        :param properties: New properties
        :return: Updated port data
        """

        return self._invoke_storage_services('SetServerPortProperties',
                                             portId=port_id,
                                             properties=properties)

    def build_access_token(self, initiator_node_name, initiator_username,
                           initiator_password, mutual_authentication,
                           target_username, target_password):
        """Create an AccessToken object.

        :param initiator_node_name: Initiator node name
        :param initiator_username: Initiator user name
        :param initiator_password: Initiator password
        :param mutual_authentication: If True the target and the initiator
                                      authenticate each other.
                                      A separate secret is set for each target
                                      and for each initiator in the storage
                                      area network (SAN).
        :param target_username: Target user name
        :param target_password: Target password
        :return: AccessToken object
        """

        access_token = (self._storage_services_client.factory
                        .create(self.O_ACCESS_TOKEN))
        access_token.InitiatorNodeName = initiator_node_name
        access_token.InitiatorUsername = initiator_username
        access_token.InitiatorPassword = initiator_password
        access_token.MutualAuthentication = mutual_authentication
        access_token.TargetUsername = target_username
        access_token.TargetPassword = target_password

        return access_token

    def set_access_token(self, iscsi_port_id, access_token):
        """Set the access token.

        The access token allows access to a specific network node
        from a specific iSCSI port.

        :param iscsi_port_id: Id of the initiator iSCSI port
        :param access_token: Access token to be validated
        :return: Port data
        """

        return self._invoke_storage_services('SetAccessToken',
                                             iScsiPortId=iscsi_port_id,
                                             inputToken=access_token)

    def get_clients(self):
        """Get all the clients in the configuration.

        :return: A list of client data
        """

        return self._get_list_data(
            self._invoke_storage_services('GetClients'),
            'ClientHostData')

    def register_client(self, host_name, description, machine_type,
                        mode, preferred_server_ids):
        """Register the client, creating a client object in the configuration.

        :param host_name: Name of the client
        :param description: Description
        :param machine_type: Type of client
        :param mode: Path policy mode of the client
        :param preferred_server_ids: Preferred server ids
        :return: New client data
        """

        client_machine_type = getattr(self._storage_services_client.factory
                                      .create(self.O_CLIENT_MACHINE_TYPE),
                                      machine_type)
        client_mode = getattr(self._storage_services_client.factory
                              .create(self.O_PATH_POLICY),
                              mode)

        return self._invoke_storage_services(
            'RegisterClient',
            hostName=host_name,
            description=description,
            type=client_machine_type,
            mode=client_mode,
            preferredServerIds=preferred_server_ids)

    def set_client_capabilities(self, client_id, mpio, alua):
        """Set the client capabilities for MPIO and ALUA.

        :param client_id: Client id
        :param mpio: If set to True then MPIO-capable
        :param alua: If set to True then ALUA-capable
        :return: Updated client data
        """

        return self._invoke_storage_services('SetClientCapabilities',
                                             clientId=client_id,
                                             mpio=mpio,
                                             alua=alua)

    def get_target_domains(self):
        """Get all the target domains in the configuration.

        :return: A list of target domains
        """

        return self._get_list_data(
            self._invoke_storage_services('GetTargetDomains'),
            'VirtualTargetDomainData')

    def create_target_domain(self, initiator_host_id, target_host_id):
        """Create a target domain given a pair of hosts, target and initiator.

        :param initiator_host_id: Id of the initiator host machine
        :param target_host_id: Id of the target host server
        :return: New target domain
        """

        return self._invoke_storage_services('CreateTargetDomain',
                                             initiatorHostId=initiator_host_id,
                                             targetHostId=target_host_id)

    def delete_target_domain(self, target_domain_id):
        """Delete a target domain.

        :param target_domain_id: Target domain id
        """

        self._invoke_storage_services('DeleteTargetDomain',
                                      targetDomainId=target_domain_id)

    def get_target_devices(self):
        """Get all the target devices in the configuration.

        :return: A list of target devices
        """

        return self._get_list_data(
            self._invoke_storage_services('GetTargetDevices'),
            'VirtualTargetDeviceData')

    def build_scsi_port_nexus_data(self, initiator_port_id, target_port_id):
        """Create a ScsiPortNexusData object.

        Nexus is a pair of ports that can communicate, one being the initiator,
        the other the target

        :param initiator_port_id: Id of the initiator port
        :param target_port_id: Id of the target port
        :return: ScsiPortNexusData object
        """

        scsi_port_nexus_data = (self._storage_services_client.factory
                                .create(self.O_SCSI_PORT_NEXUS_DATA))
        scsi_port_nexus_data.InitiatorPortId = initiator_port_id
        scsi_port_nexus_data.TargetPortId = target_port_id

        return scsi_port_nexus_data

    def create_target_device(self, target_domain_id, nexus):
        """Create a target device, given a target domain and a nexus.

        :param target_domain_id: Target domain id
        :param nexus: Nexus, or pair of ports
        :return: New target device
        """

        return self._invoke_storage_services('CreateTargetDevice',
                                             targetDomainId=target_domain_id,
                                             nexus=nexus)

    def delete_target_device(self, target_device_id):
        """Delete a target device.

        :param target_device_id: Target device id
        """

        self._invoke_storage_services('DeleteTargetDevice',
                                      targetDeviceId=target_device_id)

    def get_next_free_lun(self, target_device_id):
        """Find the next unused LUN number for a specified target device.

        :param target_device_id: Target device id
        :return: LUN number
        """

        return self._invoke_storage_services('GetNextFreeLun',
                                             targetDeviceId=target_device_id)

    def get_logical_units(self):
        """Get all the mappings configured in the system.

        :return: A list of mappings
        """

        return self._get_list_data(
            self._invoke_storage_services('GetLogicalUnits'),
            'VirtualLogicalUnitData')

    def map_logical_disk(self, logical_disk_id, nexus, lun,
                         initiator_host_id, mapping_type):
        """Map a logical disk to a host.

        :param logical_disk_id: Id of the logical disk
        :param nexus: Nexus, or pair of ports
        :param lun: Logical Unit Number
        :param initiator_host_id: Id of the initiator host machine
        :param mapping_type: Type of mapping
        :return: New mapping
        """

        logical_unit_type = getattr(self._storage_services_client.factory
                                    .create(self.O_LOGICAL_UNIT_TYPE),
                                    mapping_type)

        return self._invoke_storage_services('MapLogicalDisk',
                                             logicalDiskId=logical_disk_id,
                                             nexus=nexus,
                                             lun=lun,
                                             initiatorHostId=initiator_host_id,
                                             mappingType=logical_unit_type)

    def unmap_logical_disk(self, logical_disk_id, nexus):
        """Unmap a logical disk mapped with a specified nexus.

        :param logical_disk_id: Id of the logical disk
        :param nexus: Nexus, or pair of ports
        """

        self._invoke_storage_services('UnmapLogicalDisk',
                                      logicalDiskId=logical_disk_id,
                                      nexusData=nexus)



# Copyright (C) 2014 Deutsche Telekom AG
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.backup import driver
from cinder.tests.unit.backup import fake_service


class FakeBackupServiceWithVerify(driver.BackupDriverWithVerify,
                                  fake_service.FakeBackupService):
    def verify(self, backup):
        pass


def get_backup_driver(context):
    return FakeBackupServiceWithVerify(context)



# Copyright (c) 2016 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import oslo_serialization

from cinder.i18n import _
from cinder.volume.drivers.coprhd.helpers import commoncoprhdapi as common
from cinder.volume.drivers.coprhd.helpers import host
from cinder.volume.drivers.coprhd.helpers import project
from cinder.volume.drivers.coprhd.helpers import virtualarray
from cinder.volume.drivers.coprhd.helpers import volume


class ExportGroup(common.CoprHDResource):

    URI_EXPORT_GROUP = "/block/exports"
    URI_EXPORT_GROUPS_SHOW = URI_EXPORT_GROUP + "/{0}"
    URI_EXPORT_GROUP_SEARCH = '/block/exports/search'
    URI_EXPORT_GROUP_UPDATE = '/block/exports/{0}'

    def exportgroup_remove_volumes_by_uri(self, exportgroup_uri,
                                          volume_id_list, sync=False,
                                          tenantname=None, projectname=None,
                                          cg=None, synctimeout=0):
        """Remove volumes from the exportgroup, given the uris of volume."""

        volume_list = volume_id_list
        parms = {}

        parms['volume_changes'] = self._remove_list(volume_list)
        o = self.send_json_request(exportgroup_uri, parms)
        return self.check_for_sync(o, sync, synctimeout)

    def _remove_list(self, uris):
        resChanges = {}
        if not isinstance(uris, list):
            resChanges['remove'] = [uris]
        else:
            resChanges['remove'] = uris
        return resChanges

    def send_json_request(self, exportgroup_uri, param):
        body = oslo_serialization.jsonutils.dumps(param)
        (s, h) = common.service_json_request(
            self.ipaddr, self.port, "PUT",
            self.URI_EXPORT_GROUP_UPDATE.format(exportgroup_uri), body)
        return common.json_decode(s)

    def check_for_sync(self, result, sync, synctimeout=0):
        if sync:
            if len(result["resource"]) > 0:
                resource = result["resource"]
                return (
                    common.block_until_complete("export", resource["id"],
                                                result["id"], self.ipaddr,
                                                self.port, synctimeout)
                )
            else:
                raise common.CoprHdError(
                    common.CoprHdError.SOS_FAILURE_ERR, _(
                        "error: task list is empty, no task response found"))
        else:
            return result

    def exportgroup_list(self, project_name, tenant):
        """This function gives list of export group uris separated by comma.

        :param project_name: Name of the project path
        :param tenant: Name of the tenant
        :returns: list of export group ids separated by comma
        """
        if tenant is None:
            tenant = ""
        projobj = project.Project(self.ipaddr, self.port)
        fullproj = tenant + "/" + project_name
        projuri = projobj.project_query(fullproj)

        uri = self.URI_EXPORT_GROUP_SEARCH

        if '?' in uri:
            uri += '&project=' + projuri
        else:
            uri += '?project=' + projuri

        (s, h) = common.service_json_request(self.ipaddr, self.port, "GET",
                                             uri, None)
        o = common.json_decode(s)
        if not o:
            return []

        exportgroups = []
        resources = common.get_node_value(o, "resource")
        for resource in resources:
            exportgroups.append(resource["id"])

        return exportgroups

    def exportgroup_show(self, name, project, tenant, varray=None):
        """This function displays the Export group with details.

        :param name: Name of the export group
        :param project: Name of the project
        :param tenant: Name of the tenant
        :returns: Details of export group
        """
        varrayuri = None
        if varray:
            varrayObject = virtualarray.VirtualArray(
                self.ipaddr, self.port)
            varrayuri = varrayObject.varray_query(varray)
        uri = self.exportgroup_query(name, project, tenant, varrayuri)
        (s, h) = common.service_json_request(
            self.ipaddr,
            self.port,
            "GET",
            self.URI_EXPORT_GROUPS_SHOW.format(uri), None)
        o = common.json_decode(s)
        if o['inactive']:
            return None

        return o

    def exportgroup_create(self, name, project_name, tenant, varray,
                           exportgrouptype, export_destination=None):
        """This function creates the Export group with given name.

        :param name: Name of the export group
        :param project_name: Name of the project path
        :param tenant: Container tenant name
        :param varray: Name of the virtual array
        :param exportgrouptype: Type of the export group. Ex:Host etc
        :returns: status of creation
        """
        # check for existence of export group.
        try:
            status = self.exportgroup_show(name, project_name, tenant)
        except common.CoprHdError as e:
            if e.err_code == common.CoprHdError.NOT_FOUND_ERR:
                if tenant is None:
                    tenant = ""

                fullproj = tenant + "/" + project_name
                projObject = project.Project(self.ipaddr, self.port)
                projuri = projObject.project_query(fullproj)

                varrayObject = virtualarray.VirtualArray(
                    self.ipaddr, self.port)
                nhuri = varrayObject.varray_query(varray)

                parms = {
                    'name': name,
                    'project': projuri,
                    'varray': nhuri,
                    'type': exportgrouptype
                }

                if exportgrouptype and export_destination:
                    host_obj = host.Host(self.ipaddr, self.port)
                    host_uri = host_obj.query_by_name(export_destination)
                    parms['hosts'] = [host_uri]

                body = oslo_serialization.jsonutils.dumps(parms)
                (s, h) = common.service_json_request(self.ipaddr,
                                                     self.port, "POST",
                                                     self.URI_EXPORT_GROUP,
                                                     body)

                o = common.json_decode(s)
                return o
            else:
                raise

        if status:
            raise common.CoprHdError(
                common.CoprHdError.ENTRY_ALREADY_EXISTS_ERR, (_(
                    "Export group with name %s"
                    " already exists") % name))

    def exportgroup_query(self, name, project, tenant, varrayuri=None):
        """Makes REST API call to query the exportgroup by name.

        :param name: Name/id of the export group
        :param project: Name of the project
        :param tenant: Name of the tenant
        :param varrayuri: URI of the virtual array
        :returns: id of the export group
        """
        if common.is_uri(name):
            return name

        uris = self.exportgroup_list(project, tenant)
        for uri in uris:
            exportgroup = self.exportgroup_show(uri, project, tenant)
            if exportgroup and exportgroup['name'] == name:
                if varrayuri:
                    varrayobj = exportgroup['varray']
                    if varrayobj['id'] == varrayuri:
                        return exportgroup['id']
                    else:
                        continue
                else:
                    return exportgroup['id']
        raise common.CoprHdError(
            common.CoprHdError.NOT_FOUND_ERR,
            (_("Export Group %s: not found") % name))

    def exportgroup_add_volumes(self, sync, exportgroupname, tenantname,
                                maxpaths, minpaths, pathsperinitiator,
                                projectname, volumenames,
                                cg=None, synctimeout=0, varray=None):
        """Add volume to export group.

        :param sync: synchronous request
        :param exportgroupname: Name/id of the export group
        :param tenantname: tenant name
        :param maxpaths: Maximum number of paths
        :param minpaths: Minimum number of paths
        :param pathsperinitiator: Paths per initiator
        :param projectname: name of project
        :param volumenames: names of volumes that needs
                            to be added to exportgroup
        :param cg: consistency group
        :param synctimeout: Query for task status for 'synctimeout' secs
                            If the task doesn't complete in synctimeout secs,
                            an exception is thrown
        :param varray: Name of varray
        :returns: action result
        """
        varrayuri = None
        if varray:
            varrayObject = virtualarray.VirtualArray(
                self.ipaddr, self.port)
            varrayuri = varrayObject.varray_query(varray)

        exportgroup_uri = self.exportgroup_query(exportgroupname,
                                                 projectname,
                                                 tenantname,
                                                 varrayuri)

        # get volume uri
        if tenantname is None:
            tenantname = ""
        # List of volumes
        volume_list = []

        if volumenames:
            volume_list = self._get_resource_lun_tuple(
                volumenames, "volumes", None, tenantname,
                projectname, None)

        parms = {}
        # construct the body

        volChanges = {}
        volChanges['add'] = volume_list
        parms['volume_changes'] = volChanges

        o = self.send_json_request(exportgroup_uri, parms)
        return self.check_for_sync(o, sync, synctimeout)

    def _get_resource_lun_tuple(self, resources, resType, baseResUri,
                                tenantname, projectname, blockTypeName):
        """Function to validate input volumes and return list of ids and luns.

        """
        copyEntries = []
        volumeObject = volume.Volume(self.ipaddr, self.port)
        for copy in resources:
            copyParam = []
            try:
                copyParam = copy.split(":")
            except Exception:
                raise common.CoprHdError(
                    common.CoprHdError.CMD_LINE_ERR,
                    (_("Please provide valid format volume:"
                       " lun for parameter %s") %
                     resType))
            copy = dict()
            if not len(copyParam):
                raise common.CoprHdError(
                    common.CoprHdError.CMD_LINE_ERR,
                    (_("Please provide at least one volume for parameter %s") %
                     resType))
            if resType == "volumes":
                full_project_name = tenantname + "/" + projectname
                copy['id'] = volumeObject.volume_query(
                    full_project_name, copyParam[0])
            if len(copyParam) > 1:
                copy['lun'] = copyParam[1]
            copyEntries.append(copy)
        return copyEntries



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db.sqlalchemy import utils
from oslo_log import log as logging
from sqlalchemy import MetaData

LOG = logging.getLogger(__name__)


def ensure_index_exists(migrate_engine, table_name, column):
    index_name = table_name + '_' + column + '_idx'
    columns = [column]

    if utils.index_exists_on_columns(migrate_engine, table_name, columns):
        LOG.info(
            'Skipped adding %s because an equivalent index already exists.',
            index_name
        )
    else:
        utils.add_index(migrate_engine, table_name, index_name, columns)


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine
    for table_name, column in INDEXES_TO_CREATE:
        ensure_index_exists(migrate_engine, table_name, column)


INDEXES_TO_CREATE = (
    ('attachment_specs', 'attachment_id'),
    ('cgsnapshots', 'consistencygroup_id'),
    ('group_snapshots', 'group_id'),
    ('group_type_specs', 'group_type_id'),
    ('group_volume_type_mapping', 'group_id'),
    ('group_volume_type_mapping', 'volume_type_id'),
    ('quality_of_service_specs', 'specs_id'),
    ('reservations', 'allocated_id'),
    ('reservations', 'usage_id'),
    ('snapshot_metadata', 'snapshot_id'),
    ('snapshots', 'cgsnapshot_id'),
    ('snapshots', 'group_snapshot_id'),
    ('snapshots', 'volume_id'),
    ('transfers', 'volume_id'),
    ('volume_admin_metadata', 'volume_id'),
    ('volume_attachment', 'volume_id'),
    ('volume_glance_metadata', 'snapshot_id'),
    ('volume_glance_metadata', 'volume_id'),
    ('volume_metadata', 'volume_id'),
    ('volume_type_extra_specs', 'volume_type_id'),
    ('volume_types', 'qos_specs_id'),
    ('volumes', 'consistencygroup_id'),
    ('volumes', 'group_id'),
    ('workers', 'service_id'),
)



# Copyright (c) 2011 Zadara Storage Inc.
# Copyright (c) 2011 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The volume type & volume types extra specs extension."""

from oslo_utils import strutils
from webob import exc

from cinder.api import common
from cinder.api.openstack import wsgi
from cinder.api.v2.views import types as views_types
from cinder import exception
from cinder.i18n import _
from cinder import utils
from cinder.volume import volume_types


class VolumeTypesController(wsgi.Controller):
    """The volume types API controller for the OpenStack API."""

    _view_builder_class = views_types.ViewBuilder

    def index(self, req):
        """Returns the list of volume types."""
        limited_types = self._get_volume_types(req)
        req.cache_resource(limited_types, name='types')
        return self._view_builder.index(req, limited_types)

    def show(self, req, id):
        """Return a single volume type item."""
        context = req.environ['cinder.context']

        # get default volume type
        if id is not None and id == 'default':
            vol_type = volume_types.get_default_volume_type()
            if not vol_type:
                msg = _("Default volume type can not be found.")
                raise exception.VolumeTypeNotFound(message=msg)
            req.cache_resource(vol_type, name='types')
        else:
            # Not found  exception will be handled at wsgi level
            vol_type = volume_types.get_volume_type(context, id)
            req.cache_resource(vol_type, name='types')

        return self._view_builder.show(req, vol_type)

    def _parse_is_public(self, is_public):
        """Parse is_public into something usable.

        * True: List public volume types only
        * False: List private volume types only
        * None: List both public and private volume types
        """

        if is_public is None:
            # preserve default value of showing only public types
            return True
        elif utils.is_none_string(is_public):
            return None
        else:
            try:
                return strutils.bool_from_string(is_public, strict=True)
            except ValueError:
                msg = _('Invalid is_public filter [%s]') % is_public
                raise exc.HTTPBadRequest(explanation=msg)

    def _get_volume_types(self, req):
        """Helper function that returns a list of type dicts."""
        params = req.params.copy()
        marker, limit, offset = common.get_pagination_params(params)
        sort_keys, sort_dirs = common.get_sort_params(params)
        # NOTE(wanghao): Currently, we still only support to filter by
        # is_public. If we want to filter by more args, we should set params
        # to filters.
        filters = {}
        context = req.environ['cinder.context']
        if context.is_admin:
            # Only admin has query access to all volume types
            filters['is_public'] = self._parse_is_public(
                req.params.get('is_public', None))
        else:
            filters['is_public'] = True
        utils.remove_invalid_filter_options(context,
                                            filters,
                                            self._get_vol_type_filter_options()
                                            )
        limited_types = volume_types.get_all_types(context,
                                                   filters=filters,
                                                   marker=marker, limit=limit,
                                                   sort_keys=sort_keys,
                                                   sort_dirs=sort_dirs,
                                                   offset=offset,
                                                   list_result=True)
        return limited_types

    def _get_vol_type_filter_options(self):
        """Return volume type search options allowed by non-admin."""
        return ['is_public']


def create_resource():
    return wsgi.Resource(VolumeTypesController())
